"""
Testing configurations for the NeMoS library.

This module contains test fixtures required to set up and verify the functionality
of the modules of the NeMoS library.

Note:
    This module primarily serves as a utility for test configurations, setting up initial conditions,
    and loading predefined parameters for testing various functionalities of the NeMoS library.
"""

import abc
import os
from collections import namedtuple
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
import pynapple as nap
import pytest

import nemos as nmo
import nemos._inspect_utils as inspect_utils
import nemos.basis.basis as basis
from nemos.basis import AdditiveBasis, CustomBasis, MultiplicativeBasis
from nemos.basis._basis import Basis
from nemos.basis._basis_mixin import BasisMixin
from nemos.basis._transformer_basis import TransformerBasis

DEFAULT_KWARGS = {
    "n_basis_funcs": 5,
    "frequencies": 4,
    "window_size": 11,
    "decay_rates": np.arange(1, 1 + 5),
}

# shut-off conversion warnings
nap.nap_config.suppress_conversion_warnings = True


@pytest.fixture()
def basis_class_specific_params():
    """Returns all the params for each class."""
    all_cls = (
        list_all_basis_classes("Conv") + list_all_basis_classes("Eval") + [CustomBasis]
    )
    return {cls.__name__: cls._get_param_names() for cls in all_cls}


class BasisFuncsTesting(abc.ABC):
    """
    An abstract base class that sets the foundation for individual basis function testing.
    This class requires an implementation of a 'cls' method, which is utilized by the meta-test
    that verifies if all basis functions are properly tested.
    """

    @abc.abstractmethod
    def cls(self):
        pass


def power_func(n, x):
    return jnp.power(x, n)


def custom_basis(n_basis_funcs=5, label=None, **kwargs):
    funcs = [partial(power_func, n) for n in range(1, n_basis_funcs + 1)]
    ndim_input = kwargs.get("ndim_input", 1)
    out_shape = kwargs.get("output_shape", None)
    pynapple_support = kwargs.get("pynapple_support", True)
    return CustomBasis(
        funcs,
        label=label,
        ndim_input=ndim_input,
        output_shape=out_shape,
        pynapple_support=pynapple_support,
    )


def power_add(n, x, y):
    return x**n + y**n


def custom_basis_2d(n_basis_funcs=5, label=None):
    funcs = [partial(power_add, n) for n in range(1, n_basis_funcs + 1)]
    return CustomBasis(funcs, label=label)


def basis_collapse_all_non_vec_axis(n_basis_funcs=5, label=None, **kwargs):
    funcs = [lambda x: x.reshape(x.shape[0], -1)[:, 0] for _ in range(n_basis_funcs)]
    ndim_input = kwargs.get("ndim_input", 1)
    return CustomBasis(funcs, label=label, ndim_input=ndim_input)


def basis_with_add_kwargs(label=None, basis_kwargs=None):
    def func(x, add=0):
        return x + add

    return CustomBasis([func], label=label, basis_kwargs=basis_kwargs)


class CombinedBasis(BasisFuncsTesting):
    """
    This class is used to run tests on combination operations (e.g., addition, multiplication) among Basis functions.

    Properties:
    - cls: Class (default = None)
    """

    cls = None

    @staticmethod
    def instantiate_basis(
        n_basis, basis_class, class_specific_params, window_size=10, **kwargs
    ):
        """Instantiate and return two basis of the type specified."""

        # Set non-optional args
        new_kwargs = {
            "n_basis_funcs": n_basis,
            "window_size": window_size,
            "decay_rates": np.arange(1, 1 + n_basis),
            "frequencies": np.arange(
                (n_basis + 1) % 2, 1 + (n_basis - n_basis % 2) // 2
            ),
            "frequency_mask": None,
        }
        default_kwargs = DEFAULT_KWARGS.copy()
        default_kwargs.update(new_kwargs)
        repeated_keys = set(new_kwargs.keys()).intersection(kwargs.keys())
        if repeated_keys:
            raise ValueError(
                "Cannot set `n_basis_funcs, window_size, decay_rates` with kwargs"
            )

        # Merge with provided  extra kwargs
        kwargs = {**default_kwargs, **kwargs}

        if basis_class == AdditiveBasis:
            kwargs_mspline = inspect_utils.trim_kwargs(
                basis.MSplineEval, kwargs, class_specific_params
            )
            kwargs_raised_cosine = inspect_utils.trim_kwargs(
                basis.RaisedCosineLinearConv, kwargs, class_specific_params
            )
            b1 = basis.MSplineEval(**kwargs_mspline)
            b2 = basis.RaisedCosineLinearConv(**kwargs_raised_cosine)
            basis_obj = b1 + b2
        elif basis_class == MultiplicativeBasis:
            kwargs_mspline = inspect_utils.trim_kwargs(
                basis.MSplineEval, kwargs, class_specific_params
            )
            kwargs_raised_cosine = inspect_utils.trim_kwargs(
                basis.RaisedCosineLinearConv, kwargs, class_specific_params
            )
            b1 = basis.MSplineEval(**kwargs_mspline)
            b2 = basis.RaisedCosineLinearConv(**kwargs_raised_cosine)
            basis_obj = b1 * b2
        elif basis_class == CustomBasis:
            basis_obj = custom_basis(
                n_basis,
                **inspect_utils.trim_kwargs(basis_class, kwargs, class_specific_params),
            )
        else:
            basis_obj = basis_class(
                **inspect_utils.trim_kwargs(basis_class, kwargs, class_specific_params)
            )
        return basis_obj


# automatic define user accessible basis and check the methods
def list_all_basis_classes(filter_basis="all") -> list[BasisMixin]:
    """
    Return all the classes in nemos.basis which are a subclass of Basis,
    which should be all concrete classes except TransformerBasis.
    """
    all_basis = (
        [
            class_obj
            for _, class_obj in inspect_utils.get_non_abstract_classes(basis)
            if issubclass(class_obj, Basis)
        ]
        + [
            bas
            for _, bas in inspect_utils.get_non_abstract_classes(nmo.basis._basis)
            if bas != TransformerBasis
        ]
        + [CustomBasis]
    )
    if filter_basis != "all":
        all_basis = [a for a in all_basis if filter_basis in a.__name__]
    return all_basis


def list_all_real_basis_classes(filter_basis="all"):
    list_all_basis = list_all_basis_classes(filter_basis)
    return [cls for cls in list_all_basis if not getattr(cls, "_is_complex", False)]


# Sample subclass to test instantiation and methods
class MockRegressor(nmo.base_regressor.BaseRegressor):
    """
    Mock implementation of the BaseRegressor abstract class for testing purposes.
    Implements all required abstract methods as empty methods.
    """

    def __init__(self, std_param: int = 0):
        """Initialize a MockBaseRegressor instance with optional standard parameters."""
        self.std_param = std_param
        super().__init__()

    def fit(self, X, y):
        pass

    def predict(self, X) -> jnp.ndarray:
        pass

    def score(
        self,
        X,
        y,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def simulate(
        self,
        random_key: jax.Array,
        feed_forward_input,
        **kwargs,
    ):
        pass

    def _check_params(self, *args, **kwargs):
        pass

    def _check_input_and_params_consistency(self, *args, **kwargs):
        pass

    def _check_input_dimensionality(self, *args, **kwargs):
        pass

    def _get_coef_and_intercept(self):
        pass

    def _set_coef_and_intercept(self, params):
        pass

    def update(self, *args, **kwargs):
        pass

    def initialize_state(self, *args, **kwargs):
        pass

    def initialize_params(self, *args, **kwargs):
        pass

    def _predict_and_compute_loss(self, params, X, y):
        pass

    def _get_optimal_solver_params_config(self):
        return None, None, None

    def save_params(self, *args):
        pass


class MockRegressorNested(MockRegressor):
    def __init__(self, other_param: int, std_param: int = 0):
        super().__init__(std_param=std_param)
        self.other_param = MockGLM(std_param=other_param)


class MockGLM(nmo.glm.GLM):
    """
    Mock implementation of the BaseRegressor abstract class for testing purposes.
    Implements all required abstract methods as empty methods.
    """

    def __init__(self, std_param: int = 0):
        """Initialize a MockBaseRegressor instance with optional standard parameters."""
        self.std_param = std_param
        super().__init__()

    def fit(self, X, y):
        pass

    def predict(self, X) -> jnp.ndarray:
        pass

    def score(
        self,
        X,
        y,
        **kwargs,
    ) -> jnp.ndarray:
        pass

    def simulate(
        self,
        random_key: jax.Array,
        feedforward_input,
        **kwargs,
    ):
        pass

    def _get_coef_and_intercept(self):
        pass

    def _set_coef_and_intercept(self, params):
        pass


@pytest.fixture
def mock_regressor():
    return MockRegressor(std_param=2)


@pytest.fixture
def mock_regressor_nested():
    return MockRegressorNested(other_param=1, std_param=2)


@pytest.fixture
def mock_glm():
    return MockGLM(std_param=2)


@pytest.fixture
def poissonGLM_model_instantiation():
    """Set up a Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    observation_model = nmo.observation_models.PoissonObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer=regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate


@pytest.fixture
def poissonGLM_model_instantiation_pytree(poissonGLM_model_instantiation):
    """Set up a Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = poissonGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(model.observation_model, regularizer=model.regularizer)
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


@pytest.fixture
def poissonGLM_fitted_model_instantiation(poissonGLM_model_instantiation):
    X, y, model, true_params, firing_rate = poissonGLM_model_instantiation
    model.fit(X, y)

    return X, y, model, true_params, firing_rate


@pytest.fixture
def population_poissonGLM_model_instantiation():
    """Set up a population Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(500, 5))
    b_true = -2 * np.ones((3,))
    w_true = np.random.normal(size=(5, 3))
    observation_model = nmo.observation_models.PoissonObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model, regularizer=regularizer
    )
    rate = jnp.exp(jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate


@pytest.fixture
def population_poissonGLM_model_instantiation_pytree(
    population_poissonGLM_model_instantiation,
):
    """Set up a population Poisson GLM for testing purposes.

    This fixture initializes a Poisson GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = population_poissonGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model, regularizer=model.regularizer
    )
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


@pytest.fixture
def population_poissonGLM_fitted_model_instantiation(
    population_poissonGLM_model_instantiation,
):
    X, y, model, true_params, firing_rate = population_poissonGLM_model_instantiation
    model.fit(X, y)

    return X, y, model, true_params, firing_rate


@pytest.fixture
def coupled_model_simulate():
    """Set up a Poisson GLM from a predefined configuration in a json file.

    This fixture reads parameters for a Poisson GLM from a json configuration file, initializes
    the model accordingly, and returns the model instance with other related parameters.

    Returns:
        tuple: A tuple containing:
            - the coupling coeffs
            - the feedforward coeffs
            - the intercepts
            - jax.random.key(123) (jax.Array): A pseudo-random number generator key.
            - feedforward_input (jax.numpy.ndarray): Feedforward input values from the config.
            - coupling_basis (jax.numpy.ndarray): Coupling basis values from the config.
            - init_spikes (jax.numpy.ndarray): Initial spike values from the config.
            - a link function.

    """

    n_neurons, coupling_duration, sim_duration = 2, 100, 1000
    coupling_filter_bank = np.zeros((coupling_duration, n_neurons, n_neurons))
    for unit_i in range(n_neurons):
        for unit_j in range(n_neurons):
            coupling_filter_bank[:, unit_i, unit_j] = (
                nmo.simulation.difference_of_gammas(coupling_duration)
            )
    # shrink the filters for simulation stability
    coupling_filter_bank *= 0.8
    basis = nmo.basis.RaisedCosineLogEval(20)

    # approximate the coupling filters in terms of the basis function
    _, coupling_basis = basis.evaluate_on_grid(coupling_filter_bank.shape[0])
    coupling_coeff = nmo.simulation.regress_filter(coupling_filter_bank, coupling_basis)
    feedforward_coeff = np.ones((n_neurons, 2))
    intercepts = -3 * jnp.ones(n_neurons)

    feedforward_input = jnp.c_[
        jnp.cos(jnp.linspace(0, np.pi * 4, sim_duration)),
        jnp.sin(jnp.linspace(0, np.pi * 4, sim_duration)),
    ]

    feedforward_input = jnp.tile(feedforward_input[:, None], (1, n_neurons, 1))
    init_spikes = jnp.zeros((coupling_duration, n_neurons))

    return (
        coupling_coeff,
        feedforward_coeff,
        intercepts,
        jax.random.key(123),
        feedforward_input,
        coupling_basis,
        init_spikes,
        jnp.exp,
    )


@pytest.fixture
def jaxopt_solvers():
    return [
        "GradientDescent",
        "BFGS",
        "LBFGS",
        "ScipyMinimize",
        "NonlinearCG",
        "ScipyBoundedMinimize",
        "LBFGSB",
        "ProximalGradient",
    ]


@pytest.fixture
def poissonGLM_model_instantiation_group_sparse():
    """Set up a Poisson GLM for testing purposes with group sparse weights.

    This fixture initializes a Poisson GLM with random, group sparse, parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    w_true[1:4] = 0.0
    mask = np.zeros((2, 5))
    mask[0, 1:4] = 1
    mask[1, [0, 4]] = 1
    observation_model = nmo.observation_models.PoissonObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer=regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate, mask


@pytest.fixture
def population_poissonGLM_model_instantiation_group_sparse():
    """Set up a Poisson GLM for testing purposes with group sparse weights.

    This fixture initializes a Poisson GLM with random, group sparse, parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((3,))
    w_true = np.random.normal(size=(5, 3))
    w_true[1:4, 0] = 0.0
    mask = np.zeros((2, 5))
    mask[0, 1:4] = 1
    mask[1, [0, 4]] = 1
    observation_model = nmo.observation_models.PoissonObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(observation_model, regularizer=regularizer)
    rate = jax.numpy.exp(jax.numpy.einsum("kn,tk->tn", w_true, X) + b_true)
    return X, np.random.poisson(rate), model, (w_true, b_true), rate, mask


@pytest.fixture
def example_data_prox_operator():
    n_features = 4

    params = (
        jnp.ones((n_features)),
        jnp.zeros(
            1,
        ),
    )
    regularizer_strength = 0.1
    mask = jnp.array([[1, 0, 1, 0], [0, 1, 0, 1]]).astype(float)
    scaling = 0.5

    return params, regularizer_strength, mask, scaling


@pytest.fixture
def example_data_prox_operator_multineuron():
    n_features = 4
    n_neurons = 3

    params = (
        jnp.ones((n_features, n_neurons)),
        jnp.zeros(
            n_neurons,
        ),
    )
    regularizer_strength = 0.1
    mask = jnp.array([[1, 0, 1, 0], [0, 1, 0, 1]], dtype=jnp.float32)
    scaling = 0.5

    return params, regularizer_strength, mask, scaling


@pytest.fixture
def poisson_observation_model():
    return nmo.observation_models.PoissonObservations(jnp.exp)


@pytest.fixture
def ridge_regularizer():
    return nmo.regularizer.Ridge()


@pytest.fixture
def lasso_regularizer():
    return nmo.regularizer.Lasso(solver_name="ProximalGradient")


@pytest.fixture
def group_lasso_2groups_5features_regularizer():
    mask = np.zeros((2, 5))
    mask[0, :2] = 1
    mask[1, 2:] = 1
    return nmo.regularizer.GroupLasso(solver_name="ProximalGradient", mask=mask)


@pytest.fixture
def mock_data():
    return jnp.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]), jnp.array([[1, 2], [3, 4]])


@pytest.fixture()
def glm_class():
    return nmo.glm.GLM


@pytest.fixture()
def population_glm_class():
    return nmo.glm.PopulationGLM


@pytest.fixture
def gammaGLM_model_instantiation():
    """Set up a Gamma GLM for testing purposes.

    This fixture initializes a Gamma GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.uniform(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.uniform(size=(5,))
    observation_model = nmo.observation_models.GammaObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer=regularizer)
    rate = (jax.numpy.einsum("k,tk->t", w_true, X) + b_true) ** -1
    theta = 3
    k = rate / theta
    model.scale_ = theta
    return X, np.random.gamma(k, scale=theta), model, (w_true, b_true), rate


@pytest.fixture
def gammaGLM_model_instantiation_pytree(gammaGLM_model_instantiation):
    """Set up a Gamma GLM for testing purposes.

    This fixture initializes a Gamma GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = gammaGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(model.observation_model, regularizer=model.regularizer)
    return X_tree, spikes, model_tree, true_params_tree, rate


@pytest.fixture()
def population_gammaGLM_model_instantiation():
    np.random.seed(123)
    X = np.random.uniform(size=(500, 5))
    b_true = 0.5 * np.ones((3,))
    w_true = np.random.uniform(size=(5, 3))
    observation_model = nmo.observation_models.GammaObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model, regularizer=regularizer
    )
    rate = 1 / (jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    theta = 3
    model.scale_ = theta
    y = jax.random.gamma(jax.random.PRNGKey(123), rate / theta) * theta
    return X, y, model, (w_true, b_true), rate


@pytest.fixture()
def population_gammaGLM_model_instantiation_pytree(
    population_gammaGLM_model_instantiation,
):
    X, spikes, model, true_params, rate = population_gammaGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model, regularizer=model.regularizer
    )
    return X_tree, spikes, model_tree, true_params_tree, rate


@pytest.fixture
def regr_data():
    np.random.seed(123)
    # define inputs and coeff
    n_samples, n_features = 50, 3
    X = np.random.normal(size=(n_samples, n_features))
    coef = np.random.normal(size=(n_features))
    # set y according to lin reg eqn
    y = X.dot(coef) + 0.1 * np.random.normal(size=(n_samples,))
    return X, y, coef


@pytest.fixture
def linear_regression(regr_data):
    X, y, coef = regr_data
    # solve least-squares
    ols, _, _, _ = np.linalg.lstsq(X, y, rcond=-1)

    # set the loss
    def loss(params, X, y):
        return jnp.power(y - jnp.dot(X, params), 2).mean()

    return X, y, coef, ols, loss


@pytest.fixture
def ridge_regression(regr_data):
    X, y, coef = regr_data

    # solve least-squares
    yagu = np.hstack((y, np.zeros_like(coef)))
    Xagu = np.vstack((X, np.sqrt(0.5) * np.eye(coef.shape[0])))
    ridge, _, _, _ = np.linalg.lstsq(Xagu, yagu, rcond=-1)

    # set the loss
    def loss(params, XX, yy):
        return (
            jnp.power(yy - jnp.dot(XX, params), 2).sum()
            + 0.5 * jnp.power(params, 2).sum()
        )

    return X, y, coef, ridge, loss


@pytest.fixture
def linear_regression_tree(linear_regression):
    X, y, coef, ols, loss = linear_regression
    X_tree = dict(input_1=X[..., :2], input_2=X[..., 2:])
    coef_tree = dict(input_1=coef[:2], input_2=coef[2:])
    ols_tree = dict(input_1=ols[:2], input_2=ols[2:])

    nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, X_tree, coef_tree)

    def loss_tree(params, XX, yy):
        pred = nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, XX, params)
        return jnp.power(yy - pred, 2).sum()

    return X_tree, y, coef_tree, ols_tree, loss_tree


@pytest.fixture()
def ridge_regression_tree(ridge_regression):
    X, y, coef, ridge, loss = ridge_regression
    X_tree = dict(input_1=X[..., :2], input_2=X[..., 2:])
    coef_tree = dict(input_1=coef[:2], input_2=coef[2:])
    ridge_tree = dict(input_1=ridge[:2], input_2=ridge[2:])

    def loss_tree(params, XX, yy):
        pred = nmo.tree_utils.pytree_map_and_reduce(jnp.dot, sum, XX, params)
        norm = (
            0.5
            * nmo.tree_utils.pytree_map_and_reduce(
                lambda x: jnp.power(x, 2).sum(), sum, params
            ).sum()
        )
        return jnp.power(yy - pred, 2).sum() + norm

    return X_tree, y, coef_tree, ridge_tree, loss_tree


@pytest.fixture()
def example_X_y_high_firing_rates():
    """Example that used failed with NeMoS original initialization."""
    np.random.seed(123)

    n_features = 18
    n_neurons = 60
    n_samples = 500

    # random design array. Shape (n_time_points, n_features).
    X = 0.5 * np.random.normal(size=(n_samples, n_features))

    # log-rates & weights
    b_true = np.random.uniform(size=(n_neurons,)) * 3  # baseline rates
    w_true = np.random.uniform(size=(n_features, n_neurons)) * 0.1  # real weights:

    # generate counts (spikes will be (n_samples, n_features)
    rate = jnp.exp(jnp.dot(X, w_true) + b_true)
    spikes = np.random.poisson(rate)
    return X, spikes


@pytest.fixture
def bernoulliGLM_model_instantiation():
    """Set up a Bernoulli GLM for testing purposes.

    This fixture initializes a Bernoulli GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.binomial(1,rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.GLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    observation_model = nmo.observation_models.BernoulliObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer=regularizer)
    rate = jax.lax.logistic(jnp.einsum("k,tk->t", w_true, X) + b_true)
    return X, np.random.binomial(1, rate), model, (w_true, b_true), rate


@pytest.fixture
def bernoulliGLM_model_instantiation_pytree(bernoulliGLM_model_instantiation):
    """Set up a Bernoulli GLM for testing purposes.

    This fixture initializes a Bernoulli GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.binomial(1,rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.GLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = bernoulliGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(model.observation_model, regularizer=model.regularizer)
    return X_tree, np.random.binomial(1, rate), model_tree, true_params_tree, rate


@pytest.fixture
def population_bernoulliGLM_model_instantiation():
    """Set up a population Bernoulli GLM for testing purposes.

    This fixture initializes a Bernoulli GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.binomial(1,rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PopulationGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(500, 5))
    b_true = np.zeros((3,))
    w_true = np.random.normal(size=(5, 3))
    observation_model = nmo.observation_models.BernoulliObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model, regularizer=regularizer
    )
    rate = jax.lax.logistic(jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    return X, np.random.binomial(1, rate), model, (w_true, b_true), rate


@pytest.fixture
def population_bernoulliGLM_model_instantiation_pytree(
    population_bernoulliGLM_model_instantiation,
):
    """Set up a population Bernoulli GLM for testing purposes.

    This fixture initializes a Bernoulli GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.binomial(1,rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PopulationGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = population_bernoulliGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model, regularizer=model.regularizer
    )
    return X_tree, np.random.binomial(1, rate), model_tree, true_params_tree, rate


SizeTerminal = namedtuple("SizeTerminal", ["columns", "lines"])


class NestedRegularizer(nmo.regularizer.Ridge):
    def __init__(self, sub_regularizer, func=np.exp):
        self.sub_regularizer = sub_regularizer
        self.func = func
        super().__init__()


@pytest.fixture
def nested_regularizer():
    """
    Nested retularizer for testing save/load.
    """
    return NestedRegularizer(nmo.regularizer.Lasso())


@pytest.fixture
def negativeBinomialGLM_model_instantiation():
    """Set up a Negative Binomial GLM with array inputs for testing purposes.

    This fixture initializes a Negative Binomial GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(100, 5))
    b_true = np.zeros((1,))
    w_true = np.random.normal(size=(5,))
    observation_model = nmo.observation_models.NegativeBinomialObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.GLM(observation_model, regularizer=regularizer, solver_name="LBFGS")
    rate = jax.numpy.exp(jax.numpy.einsum("k,tk->t", w_true, X) + b_true)
    r = 1 / model.observation_model.scale
    spikes = np.random.poisson(np.random.gamma(shape=r, size=rate.shape) * (r / rate))
    return X, spikes, model, (w_true, b_true), rate


@pytest.fixture
def negativeBinomialGLM_model_instantiation_pytree(
    negativeBinomialGLM_model_instantiation,
):
    """Set up a Negative Binomial GLM with pytree inputs for testing purposes .

    This fixture initializes a Negative Binomial GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (FeaturePytree): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = negativeBinomialGLM_model_instantiation
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.GLM(
        model.observation_model, regularizer=model.regularizer, solver_name="LBFGS"
    )
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


@pytest.fixture
def population_negativeBinomialGLM_model_instantiation():
    """Set up a population Negative Binomial GLM for testing purposes.

    This fixture initializes a Negative Binomial GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    np.random.seed(123)
    X = np.random.normal(size=(500, 5))
    b_true = -2 * np.ones((3,))
    w_true = 0.1 * np.random.normal(size=(5, 3))
    observation_model = nmo.observation_models.NegativeBinomialObservations()
    regularizer = nmo.regularizer.UnRegularized()
    model = nmo.glm.PopulationGLM(
        observation_model=observation_model,
        regularizer=regularizer,
        solver_name="LBFGS",
    )
    rate = jnp.exp(jnp.einsum("ki,tk->ti", w_true, X) + b_true)
    spikes = model.observation_model.sample_generator(jax.random.PRNGKey(123), rate)
    # make sure that at least one entry is non-zero
    spikes = spikes.at[-1].set(1)
    return X, spikes, model, (w_true, b_true), rate


@pytest.fixture
def population_negativeBinomialGLM_model_instantiation_pytree(
    population_negativeBinomialGLM_model_instantiation,
):
    """Set up a population Negative Binomial GLM for testing purposes.

    This fixture initializes a Negative Binomial GLM with random parameters, simulates its response, and
    returns the test data, expected output, the model instance, true parameters, and the rate
    of response.

    Returns:
        tuple: A tuple containing:
            - X (numpy.ndarray): Simulated input data.
            - np.random.poisson(rate) (numpy.ndarray): Simulated spike responses.
            - model (nmo.glm.PoissonGLM): Initialized model instance.
            - (w_true, b_true) (tuple): True weight and bias parameters.
            - rate (jax.numpy.ndarray): Simulated rate of response.
    """
    X, spikes, model, true_params, rate = (
        population_negativeBinomialGLM_model_instantiation
    )
    X_tree = nmo.pytrees.FeaturePytree(input_1=X[..., :3], input_2=X[..., 3:])
    true_params_tree = (
        dict(input_1=true_params[0][:3], input_2=true_params[0][3:]),
        true_params[1],
    )
    model_tree = nmo.glm.PopulationGLM(
        observation_model=model.observation_model,
        regularizer=model.regularizer,
        solver_name="LBFGS",
    )
    return X_tree, np.random.poisson(rate), model_tree, true_params_tree, rate


# Select solver backend for tests if requested via environment variable
_common_solvers = {
    "SVRG": nmo.solvers.WrappedSVRG,
    "ProxSVRG": nmo.solvers.WrappedProxSVRG,
}
_solver_registry_per_backend = {
    "jaxopt": {
        **_common_solvers,
        "GradientDescent": nmo.solvers.JaxoptGradientDescent,
        "ProximalGradient": nmo.solvers.JaxoptProximalGradient,
        "LBFGS": nmo.solvers.JaxoptLBFGS,
        "BFGS": nmo.solvers.JaxoptBFGS,
        "NonlinearCG": nmo.solvers.JaxoptNonlinearCG,
    },
    "optimistix": {
        **_common_solvers,
        "GradientDescent": nmo.solvers.OptimistixOptaxGradientDescent,
        "ProximalGradient": nmo.solvers.OptimistixOptaxProximalGradient,
        "LBFGS": nmo.solvers.OptimistixOptaxLBFGS,
        "BFGS": nmo.solvers.OptimistixBFGS,
        "NonlinearCG": nmo.solvers.OptimistixNonlinearCG,
    },
}


@pytest.fixture(autouse=True, scope="session")
def configure_solver_backend():
    """
    Patch the solver registry depending on ``NEMOS_SOLVER_BACKEND``.

    Used for running solver-dependent tests in separate tox environments
    for the JAXopt and the Optimistix backends.
    """
    backend = os.getenv("NEMOS_SOLVER_BACKEND")
    if not backend:
        yield  # run with default solver registry
        return  # don't execute the remainder on teardown

    try:
        _backend_solver_registry = _solver_registry_per_backend[backend]
    except KeyError:
        available = ", ".join(_solver_registry_per_backend.keys())
        pytest.fail(f"Unknown solver backend: {backend}. Available: {available}")

    # save the original registry so that we can restore it after
    original = nmo.solvers.solver_registry.copy()
    nmo.solvers.solver_registry.clear()
    nmo.solvers.solver_registry.update(_backend_solver_registry)

    try:
        yield
    finally:
        nmo.solvers.solver_registry.clear()
        nmo.solvers.solver_registry.update(original)



from typing import Optional, Tuple

import numpy as np
from numpy._typing import ArrayLike
from numpy.typing import NDArray
from pynapple import Tsd, TsdFrame, TsdTensor

from ..type_casting import support_pynapple
from ..typing import FeatureMatrix
from ._basis import Basis, check_fraction_valid_samples, check_transform_input
from ._basis_mixin import AtomicBasisMixin
from ._check_basis import _check_input_dimensionality, _has_zero_samples


class IdentityBasis(AtomicBasisMixin, Basis):
    """
    Base class for the Identity basis.

    This basis should be used when one wants to incorporate raw inputs
    as part of a composite basis.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions.
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    """

    def __init__(
        self,
        n_basis_funcs: int,
        label: Optional[str] = None,
    ) -> None:
        AtomicBasisMixin.__init__(self, n_basis_funcs=n_basis_funcs, label=label)
        Basis.__init__(
            self,
        )
        self._n_input_dimensionality = 1

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(
        self, sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        """
        Return the samples as a 2D array.

        Parameters
        ----------
        sample_pts:
            Array of samples. The shape can be arbitrary, as long as the first axis is the
            sample axis.

        Returns
        -------
        :
            The samples with an extra axis, the n_basis_funcs axis which is = 1.

        """
        vmin = np.nanmin(sample_pts, axis=0) if self.bounds is None else self.bounds[0]
        vmax = np.nanmax(sample_pts, axis=0) if self.bounds is None else self.bounds[1]
        sample_pts[(sample_pts < vmin) | (sample_pts > vmax)] = np.nan
        check_fraction_valid_samples(
            sample_pts,
            err_msg="All the samples lie outside the [vmin, vmax] range.",
            warn_msg="More than 90% of the samples lie outside the [vmin, vmax] range.",
        )
        return sample_pts[..., np.newaxis]

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points in the uniformly spaced grid. A higher number of
            samples will result in a more detailed visualization of the basis functions.

        Returns
        -------
        X :
            Array of shape (n_samples,) containing the equi-spaced sample
            points where we've evaluated the basis.
        basis_funcs :
            The equi-spaced sample points X as a 2D array (n_samples, 1).
        """
        return super().evaluate_on_grid(n_samples)

    def _check_n_basis_min(self):
        """No checks necessary."""
        pass

    @property
    def n_basis_funcs(self) -> tuple | None:
        """Number of basis functions (read-only)."""
        return super().n_basis_funcs


class HistoryBasis(AtomicBasisMixin, Basis):
    """
    Base class for history effects.

    This basis should be used when one wants to use input history
    as a predictor.

    Parameters
    ----------
    n_basis_funcs :
        The number of basis functions (which corresponds to the length of the history window).
    label :
        The label of the basis, intended to be descriptive of the task variable being processed.
        For example: velocity, position, spike_counts.
    """

    def __init__(
        self,
        n_basis_funcs: int,
        label: Optional[str] = None,
    ) -> None:
        AtomicBasisMixin.__init__(self, n_basis_funcs=n_basis_funcs, label=label)
        Basis.__init__(
            self,
        )
        self._n_input_dimensionality = 1

    @support_pynapple(conv_type="numpy")
    @check_transform_input
    def evaluate(
        self, sample_pts: ArrayLike | Tsd | TsdFrame | TsdTensor
    ) -> FeatureMatrix:
        r"""
        Return an identity matrix of shape ``(len(samples), n_basis_funcs)``.

        The output is the convolutional kernels for spike history.

        Parameters
        ----------
        sample_pts:
            Array of samples.

        Returns
        -------
        :
            Identity matrix of shape ``(*samples.shape, n_basis_funcs)``.

        """
        sample_pts = np.squeeze(np.asarray(sample_pts))
        if sample_pts.ndim != 1:
            raise ValueError("`evaluate` for HistoryBasis allows 1D input only.")
        # this is called by set kernel.
        return np.eye(np.asarray(sample_pts).shape[0], self.n_basis_funcs)

    def evaluate_on_grid(self, n_samples: int) -> Tuple[NDArray, NDArray]:
        """Evaluate the basis set on a grid of equi-spaced sample points.

        Parameters
        ----------
        n_samples :
            The number of points used to construct the identity matrix.

        Returns
        -------
        X :
            Array of shape (n_samples,) containing a equi-spaced samples between 0 and 1.
        basis_funcs :
            The identity matrix of shape, i.e. ``np.eye(window_size, n_samples)``.
        """
        _check_input_dimensionality(self, (n_samples,))
        if _has_zero_samples((n_samples,)):
            raise ValueError("All sample counts provided must be greater than zero.")
        return np.linspace(0, 1, n_samples), np.eye(n_samples, self.n_basis_funcs)

    def _check_n_basis_min(self):
        """No checks necessary."""
        pass

    @property
    def n_basis_funcs(self) -> tuple | None:
        """Read-only property for history basis."""
        return super().n_basis_funcs



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nonlinear conjugate gradient algorithm."""

import warnings
from dataclasses import dataclass
from typing import Any, Callable, NamedTuple, Optional, Union

import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _init_stepsize
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _setup_linesearch
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype, get_real_dtype
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot_real
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_conj


class NonlinearCGState(NamedTuple):
  """Named tuple containing state information."""

  iter_num: int
  stepsize: float
  error: float
  value: float
  grad: Any
  descent_direction: Any
  aux: Optional[Any] = None

  num_fun_eval: int = 0
  num_grad_eval: int = 0
  num_linesearch_iter: int = 0


@dataclass(eq=False)
class NonlinearCG(base.IterativeSolver):
  """Nonlinear conjugate gradient solver.

  Supports complex variables, see second reference.

  Attributes:
    fun: a smooth function of the form ``fun(x, *args, **kwargs)``.
    value_and_grad: whether ``fun`` just returns the value (False) or both
      the value and gradient (True).
    has_aux: whether ``fun`` outputs auxiliary data or not.
      If ``value_and_grad == False``, the output should be
      ``value, aux = fun(...)``.
      If ``value_and_grad == True``, the output should be
      ``(value, aux), grad = fun(...)``.
      The auxiliary outputs are stored in ``state.aux``.

    maxiter: maximum number of proximal gradient descent iterations.
    tol: tolerance of the stopping criterion.

    method: which variant to calculate the beta parameter in Nonlinear CG.
      "polak-ribiere", "fletcher-reeves", "hestenes-stiefel"
      (default: "polak-ribiere")

    linesearch: the type of line search to use: "backtracking" for backtracking
      line search, "zoom" for zoom line search or "hager-zhang" for Hager-Zhang
      line search.
    linesearch_init: strategy for line-search initialization. By default, it
      will use "increase", which will increased the step-size by a factor of
      `increase_factor` at each iteration if the step-size is larger than
      `min_stepsize`, and set it to `max_stepsize` otherwise. Other choices are
      "max", that initializes the step-size to `max_stepsize` at every
      iteration, and "current", that uses the step-size from the previous
      iteration.
    condition: Deprecated. Condition used to select the stepsize when using
      backtracking linesearch.
    maxls: maximum number of iterations to use in the line search.
    decrease_factor: Deprecated. Factor by which to decrease the stepsize during
      line search when using backtracking linesearch (default: 0.8).
    increase_factor: factor by which to increase the stepsize during line search
      (default: 1.2).
    max_stepsize: upper bound on stepsize.
    min_stepsize: lower bound on stepsize guess at start of each linesearch run.
    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.

    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").

    verbose: if set to True or 1 prints the information at each step of 
      the solver, if set to 2, print also the information of the linesearch.

  References:
    Jorge Nocedal and Stephen Wright.
    Numerical Optimization, second edition.
    Algorithm 5.4 (page 121).

    Laurent Sorber, Marc van Barel, and Lieven de Lathauwer.
    Unconstrained Optimization of Real Functions in Complex Variables.
    SIAM J. Optim., Vol. 22, No. 3, pp. 879-898
  """

  fun: Callable
  value_and_grad: bool = False
  has_aux: bool = False

  maxiter: int = 100
  tol: float = 1e-3

  method: str = "polak-ribiere"  # same as SciPy
  linesearch: str = "zoom"
  linesearch_init: str = "increase"
  condition: Any = None  # deprecated in v0.8
  maxls: int = 30
  decrease_factor: Any = None  # deprecated in v0.8
  increase_factor: float = 1.2
  max_stepsize: float = 1.0
  # FIXME: should depend on whether float32 or float64 is used.
  min_stepsize: float = 1e-6

  implicit_diff: bool = True
  implicit_diff_solve: Optional[Callable] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  verbose: Union[bool, int] = False

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> NonlinearCGState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      state
    """
    (value, aux), grad = self._value_and_grad_with_aux(init_params,
                                                       *args,
                                                       **kwargs)

    dtype = tree_single_dtype(init_params)
    realdtype = get_real_dtype(dtype)

    return NonlinearCGState(iter_num=jnp.asarray(0),
                            stepsize=jnp.asarray(
        self.max_stepsize, dtype=realdtype),
        error=jnp.asarray(jnp.inf, dtype=realdtype),
        value=value,
        grad=grad,
        descent_direction=tree_scalar_mul(
        -1.0, tree_conj(grad)),
        aux=aux,
        num_fun_eval=jnp.asarray(1, base.NUM_EVAL_DTYPE),
        num_grad_eval=jnp.asarray(1, base.NUM_EVAL_DTYPE),
        num_linesearch_iter=jnp.array(
        0, base.NUM_EVAL_DTYPE)
    )

  def update(self,
             params: Any,
             state: NonlinearCGState,
             *args,
             **kwargs) -> base.OptStep:
    """Performs one iteration of Fletcher-Reeves Algorithm.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      (params, state)
    """

    eps = 1e-6
    value = state.value
    grad = state.grad
    descent_direction = state.descent_direction

    # Kept choice of no descent direction for backtracking line-search
    # FIXME: should discuss why it was the case
    if self.linesearch == "backtracking":
      ls_descent_direction = None
    else:
      ls_descent_direction = descent_direction

    init_stepsize = _init_stepsize(
        self.linesearch_init,
        self.max_stepsize,
        self.min_stepsize,
        self.increase_factor,
        state.stepsize,
    )

    new_stepsize, ls_state = self.run_ls(
        init_stepsize,
        params,
        value,
        grad,
        ls_descent_direction,
        args,
        kwargs,
    )
    new_params = ls_state.params
    new_value = ls_state.value
    new_grad = ls_state.grad
    new_aux = ls_state.aux
    new_num_fun_eval = state.num_fun_eval + ls_state.num_fun_eval
    new_num_grad_eval = state.num_grad_eval + ls_state.num_grad_eval
    new_num_linesearch_iter = state.num_linesearch_iter + ls_state.iter_num

    if self.method == "polak-ribiere":
      # See Numerical Optimization, second edition, equation (5.44).
      gTg = tree_vdot_real(grad, grad)
      gTg = jnp.where(gTg >= eps, gTg, eps)
      new_beta = tree_vdot_real(
        tree_conj(tree_sub(new_grad, grad)), tree_conj(new_grad)) / gTg
      new_beta = jax.nn.relu(new_beta)
    elif self.method == "fletcher-reeves":
      # See Numerical Optimization, second edition, equation (5.41a).
      gTg = tree_vdot_real(grad, grad)
      gTg = jnp.where(gTg >= eps, gTg, eps)
      new_beta = tree_vdot_real(new_grad, new_grad) / gTg
      new_beta = jax.nn.relu(new_beta)
    elif self.method == "hestenes-stiefel":
      # See Numerical Optimization, second edition, equation (5.45).
      grad_diff = tree_sub(new_grad, grad)
      dTg = tree_vdot_real(tree_conj(grad_diff), descent_direction)
      dTg = jnp.where(dTg >= eps, dTg, eps)
      new_beta = tree_vdot_real(
        tree_conj(grad_diff), tree_conj(new_grad)) / dTg
      new_beta = jax.nn.relu(new_beta)
    else:
      raise ValueError("method argument should be either 'polak-ribiere', "
                       "'fletcher-reeves', or 'hestenes-stiefel'.")

    new_descent_direction = tree_add_scalar_mul(tree_scalar_mul(-1, tree_conj(new_grad)),
                                                new_beta,
                                                descent_direction)
    error = tree_l2_norm(grad)
    realdtype = state.error.dtype
    new_state = NonlinearCGState(iter_num=state.iter_num + 1,
                                 stepsize=jnp.asarray(
                                   new_stepsize, dtype=realdtype),
                                 error=jnp.asarray(error, dtype=realdtype),
                                 value=new_value,
                                 grad=new_grad,
                                 descent_direction=new_descent_direction,
                                 aux=new_aux,
                                 num_fun_eval=new_num_fun_eval,
                                 num_grad_eval=new_num_grad_eval,
                                 num_linesearch_iter=new_num_linesearch_iter)

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Gradient Norm",
          additional_info={
              "Objective Value": new_value,
              "Stepsize": new_stepsize,
              "Number Linesearch Iterations": 
              new_state.num_linesearch_iter - state.num_linesearch_iter
          }
      )
    return base.OptStep(params=new_params, state=new_state)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    return self._grad_fun(params, *args, **kwargs)

  def _value_and_grad_fun(self, params, *args, **kwargs):
    (value, _), grad = self._value_and_grad_with_aux(params, *args, **kwargs)
    return value, grad

  def _grad_fun(self, params, *args, **kwargs):
    return self._value_and_grad_fun(params, *args, **kwargs)[1]

  def __post_init__(self):
    super().__post_init__()

    _fun_with_aux, _, self._value_and_grad_with_aux = base._make_funs_with_aux(
        fun=self.fun, value_and_grad=self.value_and_grad, has_aux=self.has_aux
    )

    self.reference_signature = self.fun

    unroll = self._get_unroll_option()

    linesearch_solver = _setup_linesearch(
        linesearch=self.linesearch,
        fun=_fun_with_aux,
        value_and_grad=self._value_and_grad_with_aux,
        has_aux=True,
        maxlsiter=self.maxls,
        max_stepsize=self.max_stepsize,
        jit=self.jit,
        unroll=unroll,
        verbose=max(int(self.verbose)-1, 0)
    )

    self.run_ls = linesearch_solver.run

    if self.condition is not None:
      warnings.warn("Argument condition is deprecated", DeprecationWarning)
    if self.decrease_factor is not None:
      warnings.warn(
          "Argument decrease_factor is deprecated", DeprecationWarning
      )



import pytest

import nemos._inspect_utils.inspect_utils as inspect_utils


@pytest.fixture(scope="module")
def function(request):
    func = None
    if request.param == 1:

        def func(x, *args, y=0, **kwargs):
            return

    elif request.param == 2:

        def func(x, y, *args, z=0, **kwargs):
            return

    elif request.param == 3:

        def func(x, y, z, *args, w=0, **kwargs):
            return

    return func


@pytest.mark.parametrize(
    "function, pos, var", [(1, 1, 2), (2, 2, 2), (3, 3, 2)], indirect=["function"]
)
def test_count_var_args(function, pos, var):
    num_pos, num_var = inspect_utils.count_positional_and_var_args(function)
    assert (num_pos, num_var) == (pos, var)



from typing import TYPE_CHECKING, Literal, Tuple

import jax.numpy as jnp
import numpy as np
from numpy.typing import ArrayLike, NDArray

from ..tree_utils import has_matching_axis_pytree
from ._composition_utils import infer_input_dimensionality

if TYPE_CHECKING:
    from ._basis import Basis
    from ._basis_mixin import BasisMixin


def _has_zero_samples(n_samples: Tuple[int, ...]) -> bool:
    return any([n <= 0 for n in n_samples])


def _check_zero_samples(
    n_samples, err_message="All sample provided must be non empty."
):
    if _has_zero_samples(n_samples):
        raise ValueError(err_message)


def _check_input_dimensionality(bas: "BasisMixin | Basis", xi: Tuple) -> None:
    """
    Check that the number of inputs provided by the user matches the number of inputs required.

    Parameters
    ----------
    xi[0], ..., xi[n] :
        The input samples, shape (number of samples, ).

    Raises
    ------
    ValueError
        If the number of inputs doesn't match what the Basis object requires.
    """
    n_input_dim = infer_input_dimensionality(bas)
    if len(xi) != n_input_dim:
        raise TypeError(
            f"This basis requires {n_input_dim} input(s) for evaluation, but {len(xi)} were provided."
        )


def _check_samples_consistency(*xi: NDArray) -> None:
    """
    Check that each input provided to the Basis object has the same number of time points.

    Parameters
    ----------
    xi[0], ..., xi[n] :
        The input samples, shape (number of samples, ).

    Raises
    ------
    ValueError
        If the time point number is inconsistent between inputs.
    """
    if not has_matching_axis_pytree(*xi, axis=0):
        raise ValueError(
            "Sample size mismatch. Input elements have inconsistent sample sizes."
        )


def _check_transform_input(
    bas: "BasisMixin | Basis",
    *xi: ArrayLike,
    conv_type: Literal["numpy", "jax", "none"] = "numpy",
) -> Tuple[NDArray]:
    # conversion type
    if conv_type == "numpy":
        at_least_1d = np.atleast_1d
    elif conv_type == "jax":
        at_least_1d = jnp.atleast_1d
    else:

        def at_least_1d(x):
            return x

    # check that the input is array-like (i.e., whether we can cast it to
    # numeric arrays)
    try:
        # make sure array is at least 1d (so that we succeed when only
        # passed a scalar)
        xi = tuple(at_least_1d(x).astype(float) for x in xi)
    # ValueError here surfaces the exception with e.g., `x=np.array["a", "b"])`
    except (TypeError, ValueError):
        raise TypeError("Input samples must be array-like of floats!")

    # check for non-empty samples
    _check_zero_samples(tuple(len(x) for x in xi))

    # checks on input and outputs
    _check_input_dimensionality(bas, xi)
    _check_samples_consistency(*xi)

    return xi



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loss functions."""

from typing import Callable

import jax
from jax.nn import softplus
import jax.numpy as jnp
from jax.scipy.special import logsumexp
from nemos.third_party.jaxopt.jaxopt._src.projection import projection_simplex, projection_hypercube


# Regression


def huber_loss(target: float, pred: float, delta: float = 1.0) -> float:
  """Huber loss.

  Args:
    target: ground truth
    pred: predictions
    delta: radius of quadratic behavior
  Returns:
    loss value

  References:
    https://en.wikipedia.org/wiki/Huber_loss
  """
  abs_diff = jnp.abs(target - pred)
  return jnp.where(abs_diff > delta,
                   delta * (abs_diff - .5 * delta),
                   0.5 * abs_diff ** 2)

# Binary classification.


def binary_logistic_loss(label: int, logit: float) -> float:
  """Binary logistic loss.

  Args:
    label: ground-truth integer label (0 or 1).
    logit: score produced by the model (float).
  Returns:
    loss value
  """
  # Softplus is the Fenchel conjugate of the Fermi-Dirac negentropy on [0, 1].
  # softplus = proba * logit - xlogx(proba) - xlogx(1 - proba),
  # where xlogx(proba) = proba * log(proba).
  # Use -log sigmoid(logit) = softplus(-logit)
  # and 1 - sigmoid(logit) = sigmoid(-logit).
  return softplus(jnp.where(label, -logit, logit))


def binary_sparsemax_loss(label: int, logit: float) -> float:
  """Binary sparsemax loss.

  Args:
    label: ground-truth integer label (0 or 1).
    logit: score produced by the model (float).
  Returns:
    loss value

  References:
    Learning with Fenchel-Young Losses. Mathieu Blondel, André F. T. Martins,
    Vlad Niculae. JMLR 2020. (Sec. 4.4)
  """
  return sparse_plus(jnp.where(label, -logit, logit))


def sparse_plus(x: float) -> float:
  r"""Sparse plus function.

  Computes the function:

  .. math::

    \mathrm{sparse\_plus}(x) = \begin{cases}
      0, & x \leq -1\\
      \frac{1}{4}(x+1)^2, & -1 < x < 1 \\
      x, & 1 \leq x
    \end{cases}

  This is the twin function of the softplus activation ensuring a zero output
  for inputs less than -1 and a linear output for inputs greater than 1,
  while remaining smooth, convex, monotonic by an adequate definition between
  -1 and 1.

  Args:
    x: input (float)
  Returns:
    sparse_plus(x) as defined above
  """
  return jnp.where(x <= -1.0, 0.0, jnp.where(x >= 1.0, x, (x + 1.0)**2/4))


def sparse_sigmoid(x: float) -> float:
  r"""Sparse sigmoid function.

  Computes the function:

  .. math::

    \mathrm{sparse\_sigmoid}(x) = \begin{cases}
      0, & x \leq -1\\
      \frac{1}{2}(x+1), & -1 < x < 1 \\
      1, & 1 \leq x
    \end{cases}

  This is the twin function of the sigmoid activation ensuring a zero output
  for inputs less than -1, a 1 ouput for inputs greater than 1, and a linear
  output for inputs between -1 and 1. This is the derivative of the sparse
  plus function.

  Args:
    x: input (float)
  Returns:
    sparse_sigmoid(x) as defined above
  """
  return 0.5 * projection_hypercube(x + 1.0, 2.0)


def binary_hinge_loss(label: int, score: float) -> float:
  """Binary hinge loss.

  Args:
    label: ground-truth integer label (0 or 1).
    score: score produced by the model (float).
  Returns:
    loss value.

  References:
    https://en.wikipedia.org/wiki/Hinge_loss
  """
  signed_label = 2.0 * label - 1.0
  return jnp.maximum(0, 1 - score * signed_label)


def binary_perceptron_loss(label: int, score: float) -> float:
  """Binary perceptron loss.

  Args:
    label: ground-truth integer label (0 or 1).
    score: score produced by the model (float).
  Returns:
    loss value.

  References:
    https://en.wikipedia.org/wiki/Perceptron
  """
  signed_label = 2.0 * label - 1.0
  return jnp.maximum(0, - score * signed_label)

# Multiclass classification.


def multiclass_logistic_loss(label: int, logits: jnp.ndarray) -> float:
  """Multiclass logistic loss.

  Args:
    label: ground-truth integer label, between 0 and n_classes - 1.
    logits: scores produced by the model, shape = (n_classes, ).
  Returns:
    loss value
  """
  logits = jnp.asarray(logits)
  # Logsumexp is the Fenchel conjugate of the Shannon negentropy on the simplex.
  # logsumexp = jnp.dot(proba, logits) - jnp.dot(proba, jnp.log(proba))
  # To avoid roundoff error, subtract target inside logsumexp.
  # logsumexp(logits) - logits[y] = logsumexp(logits - logits[y])
  logits = (logits - logits[label]).at[label].set(0.0)
  return logsumexp(logits)


def multiclass_sparsemax_loss(label: int, scores: jnp.ndarray) -> float:
  """Multiclass sparsemax loss.

  Args:
    label: ground-truth integer label, between 0 and n_classes - 1.
    scores: scores produced by the model, shape = (n_classes, ).
  Returns:
    loss value

  References:
    From Softmax to Sparsemax: A Sparse Model of Attention and Multi-Label
    Classification. André F. T. Martins, Ramón Fernandez Astudillo.
    ICML 2016.
  """
  scores = jnp.asarray(scores)
  proba = projection_simplex(scores)
  # Fenchel conjugate of the Gini negentropy, defined by:
  # cumulant = jnp.dot(proba, scores) + 0.5 * jnp.dot(proba, (1 - proba)).
  scores = (scores - scores[label]).at[label].set(0.0)
  return (jnp.dot(proba, jnp.where(proba, scores, 0.0))
          + 0.5 * (1.0 - jnp.dot(proba, proba)))


def multiclass_hinge_loss(label: int,
                          scores: jnp.ndarray) -> float:
  """Multiclass hinge loss.

  Args:
    label: ground-truth integer label.
    scores: scores produced by the model (floats).
  Returns:
    loss value

  References:
    https://en.wikipedia.org/wiki/Hinge_loss
  """
  one_hot_label = jax.nn.one_hot(label, scores.shape[0])
  return jnp.max(scores + 1.0 - one_hot_label) - jnp.dot(scores, one_hot_label)


def multiclass_perceptron_loss(label: int,
                               scores: jnp.ndarray) -> float:
  """Binary perceptron loss.

  Args:
    label: ground-truth integer label.
    scores: score produced by the model (float).
  Returns:
    loss value.

  References:
    Michael Collins. Discriminative training methods for Hidden Markov Models:
    Theory and experiments with perceptron algorithms. EMNLP 2002
  """
  one_hot_label = jax.nn.one_hot(label, scores.shape[0])
  return jnp.max(scores) - jnp.dot(scores, one_hot_label)

# Fenchel-Young losses


def make_fenchel_young_loss(max_fun: Callable[[jnp.ndarray], float]):
  """Creates a Fenchel-Young loss from a max function.

  Args:
    max_fun: the max function on which the Fenchel-Young loss is built.

  Returns:
    A Fenchel-Young loss function with the same signature.

  Example:
    Given a max function, e.g. the log sum exp

      from jax.scipy.special import logsumexp

      FY_loss = make_fy_loss(max_fun=logsumexp)

    Then FY loss is the Fenchel-Young loss, given for F = max_fun by

      FY_loss(y_true, scores) = F(scores) - <scores, y_true>

    Its gradient, computed automatically, is given by

      grad FY_loss = y_eps(scores) - y_true

    where y_eps is the gradient of F, the argmax.
  """

  def fy_loss(y_true, scores, *args, **kwargs):
    return max_fun(scores, *args, **kwargs) - jnp.vdot(y_true, scores)
  return fy_loss



"""Provides functionality to load a previously saved nemos model from a `.npz` file."""

import inspect
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

import numpy as np

if TYPE_CHECKING:
    from ..base_regressor import BaseRegressor
    from ..observation_models import Observations
    from ..regularizer import Regularizer

from .._observation_model_builder import (
    AVAILABLE_OBSERVATION_MODELS,
    instantiate_observation_model,
)
from .._regularizer_builder import AVAILABLE_REGULARIZERS, instantiate_regularizer
from ..glm import GLM, PopulationGLM
from ..utils import _get_name, _unflatten_dict, get_env_metadata
from ..validation import _suggest_keys

MODEL_REGISTRY = {
    "nemos.glm.glm.GLM": GLM,
    "nemos.glm.glm.PopulationGLM": PopulationGLM,
}

ERROR_MSG_OVERRIDE_NOT_ALLOWED = (
    "Cannot override the parameter {key}. "
    "NeMoS only allows overriding parameters that cannot be directly saved, "
    "such as callables, custom classes, or other objects that require pickling. "
    "If you really want to override the parameter, load the model without mapping "
    "it and then call ``set_params`` to set it afterwards."
)


def load_model(filename: Union[str, Path], mapping_dict: dict = None):
    """
    Load a previously saved nemos model from a .npz file.

    This will read the model parameters from the specified file and instantiate
    the model class with those parameters. It allows for custom mapping of
    attribute names to their actual objects using a mapping dictionary.

    Parameters
    ----------
    filename :
        Path to the saved .npz file.
    mapping_dict :
        Optional dictionary to map custom attribute names to their actual objects.

    Returns
    -------
    model :
        An instance of the model class with the loaded parameters.

    Examples
    --------
    >>> import nemos as nmo
    >>> # Create a GLM model with specified parameters
    >>> solver_args = {"stepsize": 0.1, "maxiter": 1000, "tol": 1e-6}
    >>> model = nmo.glm.GLM(
    ...     regularizer="Ridge",
    ...     regularizer_strength=0.1,
    ...     observation_model="Gamma",
    ...     solver_name="BFGS",
    ...     solver_kwargs=solver_args,
    ... )
    >>> for key, value in model.get_params().items():
    ...     print(f"{key}: {value}")
    inverse_link_function: <function one_over_x at ...>
    observation_model: GammaObservations()
    regularizer: Ridge()
    regularizer_strength: 0.1
    solver_kwargs: {'stepsize': 0.1, 'maxiter': 1000, 'tol': 1e-06}
    solver_name: BFGS
    >>> # Save the model parameters to a file
    >>> model.save_params("model_params.npz")
    >>> # Load the model from the saved file
    >>> model = nmo.load_model("model_params.npz")
    >>> # Model has the same parameters before and after load
    >>> for key, value in model.get_params().items():
    ...     print(f"{key}: {value}")
    inverse_link_function: <function one_over_x at ...>
    observation_model: GammaObservations()
    regularizer: Ridge()
    regularizer_strength: 0.1
    solver_kwargs: {'stepsize': 0.1, 'maxiter': 1000, 'tol': 1e-06}
    solver_name: BFGS

    >>> # Loading a custom inverse link function
    >>> model = nmo.glm.GLM(inverse_link_function=lambda x: x**2)
    >>> model.save_params("model_params.npz")
    >>> # Provide a mapping for the custom link function when loading.
    >>> mapping_dict = {
    ...     "inverse_link_function": lambda x: x**2,
    ... }
    >>> loaded_model = nmo.load_model("model_params.npz", mapping_dict=mapping_dict)
    >>> # Now the loaded model will have the updated solver_name and solver_kwargs
    >>> for key, value in loaded_model.get_params().items():
    ...     print(f"{key}: {value}")
    inverse_link_function: <function <lambda> at ...>
    observation_model: PoissonObservations()
    regularizer: UnRegularized()
    regularizer_strength: None
    solver_kwargs: {}
    solver_name: GradientDescent
    """
    # load the model from a .npz file
    filename = Path(filename)
    data = np.load(filename, allow_pickle=False)

    # unflatten dictionary
    saved_params = _unflatten_dict(data)
    # "save_metadata" is used to store versions of Nemos and Jax, not needed for loading
    saved_params.pop("save_metadata")
    # unflatten user map
    nested_map_dict, key_not_found = _unflattened_user_map(mapping_dict, saved_params)

    invalid_keys = _get_invalid_mappings(nested_map_dict)
    if len(invalid_keys) > 0:
        raise ValueError(
            "Invalid map parameter types detected. "
            f"The following parameters are non mappable:\n\t{invalid_keys}\n"
            "Only callables and classes can be mapped."
        )

    # backtrack all errors
    if key_not_found:
        available_keys = get_user_keys_from_nested_dict(saved_params)
        requested_keys = get_user_keys_from_nested_dict(mapping_dict)
        not_available = sorted(set(requested_keys).difference(available_keys))
        suggested_pairs = _suggest_keys(not_available, available_keys)
        suggestions = "".join(
            [
                (
                    f"\t- '{provided}', did you mean '{suggested}'?\n"
                    if suggested is not None
                    else f"\t- '{provided}'\n"
                )
                for provided, suggested in suggested_pairs
            ]
        )
        raise ValueError(
            "The following keys in your mapping do not match any parameters in the loaded model:\n\n"
            f"{suggestions}\n"
            "Please double-check your mapping dictionary."
        )
    # if any value from saved_params is a key in mapping_dict,
    # replace it with the corresponding value from mapping_dict
    saved_params, updated_keys = _apply_custom_map(saved_params, nested_map_dict)

    if len(updated_keys) > 0:
        warnings.warn(
            f"The following keys have been replaced in the model parameters: {updated_keys}.",
            UserWarning,
        )

    # Extract the model class from the saved attributes
    model_name = str(saved_params.pop("model_class"))
    model_class = MODEL_REGISTRY[model_name]

    config_params, fit_params = _split_model_params(saved_params, model_class)

    # Create the model instance
    try:
        model = model_class(**config_params)
    except Exception:
        raise ValueError(
            f"Failed to instantiate model class '{model_name}' with parameters: {config_params}. "
            f"Use `nmo.inspect_npz('{filename}')` to inspect the saved object."
        )

    # Set the rest of the parameters as attributes if recognized
    _set_fit_params(model, fit_params, filename)

    return model


def _is_param(par):
    if not isinstance(par, dict):
        return True
    return "class" not in par


def _safe_instantiate(
    parameter_name: str, class_name: str, **kwargs
) -> "Regularizer | Observations":
    if not isinstance(class_name, str):
        # this should not be hit, if it does the saved params had been modified.
        raise ValueError(
            f"Parameter ``{parameter_name}`` cannot be initialized. "
            "When a parameter specifies a class, the class name must be a string. "
            f"Class name for the loaded parameter is {class_name}."
        )
    class_basename = class_name.split(".")[-1]
    if class_basename in AVAILABLE_REGULARIZERS:
        return instantiate_regularizer(class_name, **kwargs)
    elif any(class_basename.startswith(obs) for obs in AVAILABLE_OBSERVATION_MODELS):
        return instantiate_observation_model(class_name, **kwargs)
    else:
        # Hit when loading a custom class without mapping
        if parameter_name == "observation_model":
            class_type = "observation"
        else:
            class_type = "regularization"
        raise ValueError(
            f"The class '{class_basename}' is not a native NeMoS class.\n"
            f"To load a custom {class_type} class, please provide the following mapping:\n\n"
            f" - nemos.load_model(save_path, mapping_dict={{'{parameter_name}': {class_basename}}})"
        )


def _apply_custom_map(
    params: dict, mapping_dict: dict, updated_keys: List | None = None
) -> Tuple[dict, List]:
    """
    Recursively apply user-defined mappings to a saved parameter structure.

    This function processes the nested parameter dictionary produced by `_unflatten_dict`
    and applies user-specified overrides where allowed. It does the following:

    - For leaf parameters stored as [value, [is_mapped, mapped_value]]:
        * If `is_mapped` is True, the parameter is replaced with `mapped_value`.
          Only callables or classes are allowed; other types raise an error.
        * If `is_mapped` is False, the original saved value is kept.
    - For nested dictionaries of parameters (e.g., solver kwargs):
        * These cannot be overridden because they are not callables or classes.
        * All leaves are recursively unwrapped to extract the original saved values,
          discarding any mapping info.
    - For parameters representing classes:
        * If not mapped, the original class name is checked and instantiated safely using `_safe_instantiate`.
        * If mapped, the mapping must be an actual Python class object (not a string or an instance).
          This invariant is enforced with an internal assertion for developer safety.

    This function also keeps track of which keys were overridden by the user-supplied mapping,
    returning this list alongside the reconstructed parameter dictionary.

    Parameters
    ----------
    params :
        The nested saved parameters to process. Each entry is either:
          - A leaf in the form [value, [is_mapped, mapped_value]], or
          - A nested dict representing classes.
    mapping_dict:
        A dict of mappings following the same keypath that params follows (a nested dict).
    updated_keys :
        List of keys that have already been updated, used for accumulating changes
        across recursive calls.

    Returns
    -------
    updated_params :
        The new parameter dictionary with mappings applied and wrappers removed.
    updated_keys :
        List of all keys that were actually overridden.

    Raises
    ------
    ValueError
        If a user tries to override a parameter with an unsupported type (non-callable, non-class),
        or provides a mapped class as a string instead of a Python class object. This is triggered
        in `_safe_instantiate`.

    Notes
    -----
    This function enforces strict override safety: only callables and classes may be
    mapped at load time. Directly serializable values and nested dictionaries cannot
    be overridden and must be changed later using `set_params` if needed.

    Internal invariants are checked with `assert` to ensure that only valid class mappings
    reach instantiation. If these assertions fail, it indicates a bug in the input validation
    logic and should never occur in normal use.
    """
    updated_params = {}

    if updated_keys is None:
        updated_keys = []

    for key, val in params.items():
        # handle classes and params separately
        if _is_param(val):
            if isinstance(val, dict):
                # dict cannot be mapped, so store original params
                updated_params[key] = val
            else:
                mapped_val = mapping_dict.get(key, None)
                is_mapped = mapped_val is not None
                if is_mapped:
                    updated_params[key] = mapped_val
                    updated_keys.append(key)
                else:
                    updated_params[key] = val

        else:
            # if val is a class, it must be a dict with a "class" key
            class_name = val.pop("class")
            mapped_val = mapping_dict.get(key, {})
            is_mapped = "class" in mapped_val
            mapped_params = mapped_val.get("params", {})
            if not is_mapped:
                # check for nested callable/classes save instantiate based on the string
                new_params, updated_keys = _apply_custom_map(
                    val.pop("params", {}), mapped_params, updated_keys=updated_keys
                )
                updated_params[key] = _safe_instantiate(key, class_name, **new_params)
            else:
                mapped_class = mapped_val["class"]
                updated_keys.append(key)
                # Should not be hit ever, assertion for developers
                assert inspect.isclass(mapped_class), (
                    f"The parameter '{key}' passed the type check in "
                    f"``nmo.load_model`` but is not callable or class, "
                    "check why this is the case."
                )
                # map callables and nested classes
                new_params, updated_keys = _apply_custom_map(
                    val.pop("params", {}), mapped_params, updated_keys=updated_keys
                )
                # try instantiating it with the params
                # this executes code, but we are assuming that the mapped_class is safe
                updated_params[key] = mapped_class(**new_params)

    return updated_params, updated_keys


def _split_model_params(params: dict, model_class) -> tuple:
    """Split parameters into config and fit parameters."""
    model_param_names = model_class._get_param_names()
    config_params = {k: v for k, v in params.items() if k in model_param_names}
    fit_params = {k: v for k, v in params.items() if k not in model_param_names}
    return config_params, fit_params


def _set_fit_params(model: "BaseRegressor", fit_params: dict, filename: Path):
    """Set fit model attributes, warn if unrecognized."""
    check_str = (
        f"\nIf this is confusing, try calling "
        f"`{_get_name(inspect_npz)}('{filename}')` to inspect the saved object."
    )
    for key, value in fit_params.items():
        if hasattr(model, key):
            setattr(model, key, value)
        else:
            raise ValueError(
                f"Unrecognized attribute '{key}' during model loading.{check_str}"
            )


def inspect_npz(file_path: Union[str, Path]):
    """
    Examine a saved model parameter file (.npz).

    Prints out all keys and associated values.

    Parameters
    ----------
    file_path :
        Path to the `.npz` file containing the saved model parameters.
    """
    file_path = Path(file_path)
    data = np.load(file_path, allow_pickle=True)
    data = _unflatten_dict(data)

    pad_len = max(len(k) for k in data.keys()) + 2

    metadata: dict | None = data.pop("save_metadata", None)
    installed_env = get_env_metadata()
    if metadata is not None:
        print("Metadata\n--------")
        for k, v in metadata.items():
            label = f"{k} version"
            print(f"{label:<{pad_len}}: {v}" f" (installed: {installed_env[k]})")

    print("\nModel class\n-----------")
    model_class = data.pop("model_class", None)
    if model_class:
        print(f"{'Saved model class':<{pad_len}}: {model_class}")

    print("\nModel parameters\n----------------")
    config_params = {k: data.pop(k) for k in list(data) if not k.endswith("_")}
    for key in config_params:
        val = config_params[key]
        # If the value is a callable, print its name, otherwise print the value
        if hasattr(val, "__name__"):
            print(f"{key:<{pad_len}}: {_get_name(val)}")
        else:
            print(f"{key:<{pad_len}}: {val}")

    print("\nModel fit parameters\n--------------------")
    for param in data:
        print(f"{param}: {data[param]}")


def _unflattened_user_map(
    mapping_dict: dict | None, nested_dict: dict
) -> Tuple[dict, bool]:
    """Expand user key mapping path to match saved keys."""
    if mapping_dict is None:
        return {}, False
    nested_mapping = {}
    for user_key, mapped_value in mapping_dict.items():
        current = nested_mapping
        subdict = nested_dict
        parts = [k[::-1] for k in user_key[::-1].split("__")][::-1]
        for part in parts[:-1]:
            if part not in subdict:
                return {}, True
            subdict = nested_dict[part]
            if part not in current:
                current[part] = {}
            current = current[part]
            # in we hit this case, this is a parameter (this is not the last
            # part of the key, but the class is always the last element)
            if isinstance(subdict, dict) and "class" in subdict:
                subdict = subdict["params"]
                if "params" not in current:
                    current["params"] = {}
                current = current["params"]

        if parts[-1] not in subdict:
            return None, True
        elif isinstance(subdict[parts[-1]], dict) and "class" in subdict[parts[-1]]:
            current[parts[-1]] = {"class": mapped_value}
        else:
            current[parts[-1]] = mapped_value
    return nested_mapping, False


def get_user_keys_from_nested_dict(nested_dict: dict, filter_keys: bool = True) -> list:
    """
    Get the user-formatted keys from a nested dictionary.

    Retrieve user-formatted keys from a nested parameter dictionary. The formatting matches the
    sklearn-style parameter naming convention (e.g., 'regularizer__solver_name'). This format
    should be used when providing a ``mapping_dict`` to ``load_model`` to override saved parameters.

    Parameters
    ----------
    nested_dict :
        A nested parameter dictionary, typically from a saved model.
    filter_keys :
        If True, remove internal keys ('__class' and '__params') from the output and return
        only user-facing parameter names. Default is True.

    Returns
    -------
    list of str
        A sorted list of user-formatted keys in sklearn parameter style, where nested attributes
        are joined with double underscores (e.g., 'observation_model__class', 'solver_kwargs__tol').
        The 'save_metadata' key is excluded from the output.

    Notes
    -----
    - Keys are formatted using double underscores ('__') as separators to match sklearn conventions
    - The 'save_metadata' key is automatically excluded from results
    - Internal structure keys ('__params') are filtered out when filter_keys=True

    Examples
    --------
    >>> params = {
    ...     'regularizer': {'class': 'GroupLasso', 'params': {'mask': None}},
    ...     'solver_kwargs': {'tol': 1e-7, 'maxiter': 100}
    ... }
    >>> get_user_keys_from_nested_dict(params)
    ['regularizer', 'regularizer__mask', 'solver_kwargs', 'solver_kwargs__maxiter', 'solver_kwargs__tol']
    """
    valid_keys = list(nested_dict.keys())
    sep = "__"
    for key in nested_dict.keys():
        if isinstance(nested_dict[key], dict):
            new_keys = get_user_keys_from_nested_dict(nested_dict[key], False)
            valid_keys.extend([key + sep + new for new in new_keys if new != "class"])
    if filter_keys:
        valid_keys = sorted(list({re.sub("(__params)", "", key) for key in valid_keys}))
    return valid_keys


def _get_invalid_mappings(mapping_dict: dict | None) -> List:
    """
    Recursively identify invalid entries in a model mapping dictionary.

    Validate a nested mapping dictionary by collecting keys whose values are not classes, callables,
    or valid nested mappings. This function is used during model deserialization (e.g., in
    ``nmo.load_model``) to validate a user-provided ``mapping_dict`` that maps
    symbolic model components to actual Python objects such as classes,
    callables, or parameter specifications.

    A mapping entry is considered **valid** if:
        * Its value is a class (``inspect.isclass(v)``), or
        * Its value is callable (e.g., a function or lambda), or
        * It is a dictionary containing a ``"class"`` key — in which case all
          of its subparameters (e.g., under ``"params"``) are accepted without
          validation, since they are assumed to be constructor arguments for
          the specified class.
        * It is a dictionary containing only a ``"params"`` key, in which case
          the function recursively validates each entry in ``v["params"]``.

    Any entry that does not meet these criteria is considered **invalid** and
    its key (or nested key path) is returned.

    Parameters
    ----------
    mapping_dict :
        A (possibly nested) dictionary defining how symbolic model components
        should be mapped to Python classes or callables. May contain nested
        entries with the special keys ``"class"`` and/or ``"params"``.

    Returns
    -------
    list of str
        A list of invalid key paths. For nested invalid entries, keys are joined
        with double underscores (``"__"``) to indicate hierarchy, e.g.,
        ``"regularizer__mask"``.

    Notes
    -----
    This function allows a ``"class"`` entry to bypass validation of its
    parameters because those parameters are intended to be passed to the class
    constructor during model instantiation.

    Examples
    --------
    >>> import nemos as nmo
    >>> def square(x):
    ...     return x**2
    ...
    >>> class MyRegularizer(nmo.regularizer.GroupLasso):
    ...     def __init__(self, mask=None, new_param=1):
    ...         super().__init__(mask=mask)
    ...         self.new_param = new_param
    ...
    >>> invalid = _get_invalid_mappings({
    ...     "inverse_link_function": square,
    ...     "regularizer": {"class": MyRegularizer, "params": {"regularizer__new_param": 10.}}
    ... })
    >>> invalid
    []
    >>> invalid = _get_invalid_mappings({
    ...     "inverse_link_function": square,
    ...     "regularizer": {"params": {"new_param": 10.}}
    ... })
    >>> invalid
    ['regularizer__new_param']
    """
    if mapping_dict is None:
        return []

    invalid = []

    for key, v in mapping_dict.items():
        if isinstance(v, dict) and "class" in v:
            if inspect.isclass(v["class"]):
                # Valid class - skip all validation including params
                continue
            else:
                # Invalid class - mark it
                invalid.append(key)
                # But still validate params if present, for completeness in the error message
                if "params" in v:
                    invalid_sub = _get_invalid_mappings(v["params"])
                    invalid.extend(f"{key}__{k}" for k in invalid_sub)
        # Handle dict with "params" key (but no "class" or already processed)
        elif isinstance(v, dict) and "params" in v:
            invalid_sub = _get_invalid_mappings(v["params"])
            invalid.extend(f"{key}__{k}" for k in invalid_sub)
        # Handle non-dict values
        elif not inspect.isclass(v) and not callable(v):
            invalid.append(key)

    return invalid



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import projection
from nemos.third_party.jaxopt.jaxopt.base import KKTSolution
from nemos.third_party.jaxopt.jaxopt import EqualityConstrainedQP
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class EqualityConstrainedQPTest(test_util.JaxoptTestCase):
  def _check_derivative_Q_c_A_b(self, solver, params, Q, c, A, b):
    def fun(Q, c, A, b):
      Q = 0.5 * (Q + Q.T)

      hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))

      # reduce the primal variables to a scalar value for test purpose.
      return jnp.sum(solver.run(**hyperparams).params[0])

    # Derivative w.r.t. A.
    rng = onp.random.RandomState(0)
    V = rng.rand(*A.shape)
    V /= onp.sqrt(onp.sum(V ** 2))
    eps = 1e-4
    deriv_jax = jnp.vdot(V, jax.grad(fun, argnums=2)(Q, c, A, b))
    deriv_num = (fun(Q, c, A + eps * V, b) - fun(Q, c, A - eps * V, b)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, atol=1e-3)

    # Derivative w.r.t. b.
    v = rng.rand(*b.shape)
    v /= onp.sqrt(onp.sum(v ** 2))
    eps = 1e-4
    deriv_jax = jnp.vdot(v, jax.grad(fun, argnums=3)(Q, c, A, b))
    deriv_num = (fun(Q, c, A, b + eps * v) - fun(Q, c, A, b - eps * v)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, atol=1e-3)

    # Derivative w.r.t. Q
    W = rng.rand(*Q.shape)
    W /= onp.sqrt(onp.sum(W ** 2))
    eps = 1e-4
    deriv_jax = jnp.vdot(W, jax.grad(fun, argnums=0)(Q, c, A, b))
    deriv_num = (fun(Q + eps * W, c, A, b) - fun(Q - eps * W, c, A, b)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, atol=1e-3)

    # Derivative w.r.t. c
    w = rng.rand(*c.shape)
    w /= onp.sqrt(onp.sum(w ** 2))
    eps = 1e-4
    deriv_jax = jnp.vdot(w, jax.grad(fun, argnums=1)(Q, c, A, b))
    deriv_num = (fun(Q, c + eps * w, A, b) - fun(Q, c - eps * w, A, b)) / (2 * eps)
    self.assertAllClose(deriv_jax, deriv_num, atol=1e-3)

  def test_qp_eq_only(self):
    Q = 2 * jnp.array([[2.0, 0.5], [0.5, 1]])
    c = jnp.array([1.0, 1.0])
    A = jnp.array([[1.0, 1.0]])
    b = jnp.array([1.0])
    qp = EqualityConstrainedQP(tol=1e-7)
    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))
    sol = qp.run(**hyperparams).params
    self.assertAllClose(qp.l2_optimality_error(sol, **hyperparams), 0.0)
    self._check_derivative_Q_c_A_b(qp, hyperparams, Q, c, A, b)

  def test_qp_eq_with_init(self):
    Q = 2 * jnp.array([[2.0, 0.5], [0.5, 1]])
    c = jnp.array([1.0, 1.0])
    A = jnp.array([[1.0, 1.0]])
    b = jnp.array([1.0])
    qp = EqualityConstrainedQP(tol=1e-7)
    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))
    init_params = KKTSolution(jnp.array([1.0, 1.0]), jnp.array([1.0]))
    sol = qp.run(init_params, **hyperparams).params
    self.assertAllClose(qp.l2_optimality_error(sol, **hyperparams), 0.0)
    self._check_derivative_Q_c_A_b(qp, hyperparams, Q, c, A, b)

  def test_projection_hyperplane(self):
    x = jnp.array([1.0, 2.0])
    a = jnp.array([-0.5, 1.5])
    b = 0.3
    # Find ||y-x||^2 such that jnp.dot(y, a) = b.
    expected = projection.projection_hyperplane(x, (a, b))

    matvec_Q = lambda params_Q, u: u
    matvec_A = lambda params_A, u: jnp.dot(a, u).reshape(1)
    qp = EqualityConstrainedQP(matvec_Q=matvec_Q, matvec_A=matvec_A)
    # In this example, params_Q = params_A = None.
    hyperparams = dict(params_obj=(None, -x), params_eq=(None, jnp.array([b])))
    sol = qp.run(**hyperparams).params
    primal_sol = sol[0]
    self.assertArraysAllClose(primal_sol, expected)
    self.assertAllClose(qp.l2_optimality_error(sol, **hyperparams), 0.0)

  def test_pytree_api(self):
    Q1 = jnp.array([[1.0, -0.5],
                  [-0.5, 1.0]])
    Q2 = jnp.array([[2.0]])
    Q = {'problem1': Q1, 'problem2': Q2}

    c1 = jnp.array([-0.4, 0.3])
    c2 = jnp.array([0.1])
    c = {'problem1': c1, 'problem2': c2}

    a1 = jnp.array([[-0.5, 1.5]])
    a2 = jnp.array([[10.0]])
    A = {'problem1': a1, 'problem2': a2}

    b1 = jnp.array([0.3])
    b2 = jnp.array([5.0])
    b = {'problem1': b1, 'problem2': b2}

    qp = EqualityConstrainedQP(tol=1e-3)
    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))
    # Test that there is no error when running the solver.
    sol = qp.run(**hyperparams).params

  @parameterized.product(refine_regularization=[0., 1e-6])
  def test_matvec_api(self, refine_regularization):
    # refine_regularization != 0. triggers a non regression test for issue #311
    rng = onp.random.RandomState(0)
    Q = rng.randn(7, 7)
    Q = onp.dot(Q, Q.T)
    A = rng.randn(4, 7)

    tmp = rng.randn(7)
    # Must have the same pytree structure as the output of matvec_Q.
    c = (tmp[:3], tmp[3:])
    # Must have the same pytree structure as the output of matvec_A.
    b = rng.randn(4)

    def matvec_Q(Q, tup):
      x_ = jnp.concatenate(tup)
      res = jnp.dot(Q, x_)
      return res[:3], res[3:]

    def matvec_A(A, tup):
      x_ = jnp.concatenate(tup)
      return jnp.dot(A, x_)

    # With pytrees directly.
    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))
    qp = EqualityConstrainedQP(matvec_Q=matvec_Q, matvec_A=matvec_A,
                               refine_regularization=refine_regularization)
    # sol.primal has the same pytree structure as the output of matvec_Q.
    # sol.dual_eq has the same pytree structure as the output of matvec_A.
    sol_pytree = qp.run(**hyperparams).params
    self.assertAllClose(
      qp.l2_optimality_error(sol_pytree, **hyperparams), 0.0, atol=1e-4
    )

    # With flattened pytrees.
    hyperparams = dict(params_obj=(Q, jnp.concatenate(c)), params_eq=(A, b))
    qp = EqualityConstrainedQP()
    sol = qp.run(**hyperparams).params
    self.assertAllClose(qp.l2_optimality_error(sol, **hyperparams), 0.0, atol=1e-4)

    # Check that the solutions match.
    self.assertArraysAllClose(jnp.concatenate(sol_pytree.primal), sol.primal, atol=1e-4)
    self.assertArraysAllClose(sol_pytree.dual_eq, sol.dual_eq, atol=1e-4)

  def test_fun_api(self):
    # mimic a function with pytree parameters.
    x = {'complicated_pytree': jnp.array([1.0, 2.0])}
    a = jnp.array([-0.5, 1.5])
    b = jnp.array(0.3)

    # convex quadratic and pytree parameters.
    def euclidean_quadratic_cost(y, params_obj):
      x = params_obj['complicated_pytree']
      return jnp.sum((x - y)*(x - y))
    
    def matvec_A(params_A, u):
      return jnp.dot(params_A, u).reshape(1)
    
    qp = EqualityConstrainedQP(fun=euclidean_quadratic_cost,
                               matvec_A=matvec_A)    

    try:
      # attempt to run without providing init_params, despite fun != None.
      hyperparams = dict(params_obj=x, params_eq=(a, b))
      sol = qp.run(**hyperparams).params
    except ValueError as e:
      pass  # here, a failure is the expected behavior.
    else:
      self.fail("Expected ValueError when init_params is not provided (`fun` API).")

    def scalar_solve_qp(hyperparams, init_params):
      params = qp.run(init_params, **hyperparams).params
      # arbitrary scalar function to ensure that the gradient is defined.
      return jnp.sum(params.primal)  

    hyperparams = dict(params_obj=x, params_eq=(a, jnp.array([b])))
    init_params = KKTSolution(jnp.array([1.0, 1.0]), jnp.array([1.0]))

    solve_value_grad = jax.value_and_grad(scalar_solve_qp)
    # we only need to test if the gradient is well defined in reverse mode AD:
    sol, unused_grad = solve_value_grad(hyperparams, init_params)

  def test_example_from_doc(self):
    """Test the example exposed in quadratic_problems.rst"""
    Q = jnp.array([[1.0, 0.5], [0.5, 4.0]])
    c = jnp.zeros(Q.shape[0])
    A = jnp.array([[1.0, 2.0]])
    b = jnp.array([1.0])

    def wrong_solver(Q):  # don't do this!

      def matvec_Q(params_Q, x):
        del params_Q  # unused
        # error! Q is captured from the global scope.
        # it does not fail now, but it will later.
        return jnp.dot(Q, x)
      
      eq_qp = EqualityConstrainedQP(matvec_Q=matvec_Q)
      # Warning: Q does not appear in the call to run()
      sol = eq_qp.run(None, params_obj=(None, c), params_eq=(A, b)).params
      loss = jnp.sum(sol.primal)
      return loss

    _ = wrong_solver(Q)  # no error... but it will fail later.
    try:
      _ = jax.grad(wrong_solver)(Q)  # error!
    except jax._src.interpreters.ad.CustomVJPException as e:
      pass  # here, a failure is the expected behavior.
    else:
      self.fail("Expected CustomVJPException when matvec_Q captures Q.")

    def correct_solver(Q):

      def matvec_Q(params_Q, x):
        return jnp.dot(params_Q, x)

      eq_qp = EqualityConstrainedQP(matvec_Q=matvec_Q)
      # Q is passed as a parameter, not captured from the global scope.
      sol = eq_qp.run(None, params_obj=(Q, c), params_eq=(A, b)).params
      loss = jnp.sum(sol.primal)
      return loss

    _ = correct_solver(Q)  # no error
    _ = jax.grad(correct_solver)(Q)  # no error

  def test_challenging_problem(self):
    size_prob = 100
    size_eq = 98

    onp.random.seed(511)
    X = onp.random.rand(size_prob, size_prob)
    Q = jnp.array(X @ X.T)
    c = jnp.array(onp.random.rand(size_prob))

    A = jnp.array(onp.random.rand(size_eq, size_prob))
    b = jnp.array(onp.random.rand(size_eq))

    # The tolerance is low, but ordinarily this problem cannot be solved by
    # EqualityConstrainedQP without refinement, because the matrix has a huge
    # condition number.
    low_tol = 1e-1
    qp = EqualityConstrainedQP(
      refine_regularization=1.0,
      refine_maxiter=3000,
      tol=low_tol,
      jit=True,
    )

    hyperparams = dict(params_obj=(Q, c), params_eq=(A, b))
    sol, state = qp.run(**hyperparams)
    error = qp.l2_optimality_error(sol, **hyperparams)
    self.assertLess(error, low_tol)


if __name__ == "__main__":
  absltest.main()



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def plot_categorical_var_design_matrix():
    # Data
    data = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
            [0.0, 1.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 1.0],
        ]
    )

    plt.figure(figsize=(6, 6))
    ax = sns.heatmap(
        data,
        annot=True,
        fmt=".0f",
        cmap="Pastel1_r",
        cbar=False,
        linewidths=0.5,
        linecolor="black",
        annot_kws={"fontsize": 22},
    )
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    ax.spines["bottom"].set_visible(True)
    ax.spines["right"].set_visible(True)
    # Show plot
    plt.tight_layout(pad=2.0)


def generate_lfp_like_signal(duration, sampling_rate, seed):
    """
    Generate a synthetic LFP-like signal.

    Parameters:
    duration (float): Duration of the signal in seconds.
    sampling_rate (int): Sampling rate in Hz.

    Returns:
    np.ndarray: LFP-like signal.
    """
    np.random.seed(seed)  # For reproducibility

    # Time array
    t = np.linspace(0, duration, int(duration * sampling_rate))

    # Generate power spectrum with realistic parameters
    freqs = np.fft.rfftfreq(len(t), 1 / sampling_rate)

    # Power-law decay
    alpha = 1.0  # Moderate decay exponent for LFP signals
    power_spectrum = 10**3 / (freqs + 1e-6) ** alpha

    # Add a realistic peak at ~5 Hz (e.g., theta band)
    peak_freq = 5  # Hz
    peak_width = 2  # Width of the peak in Hz
    power_spectrum += np.exp(-0.5 * ((freqs - peak_freq) / peak_width) ** 2) * 300

    # Smooth transition to higher frequencies
    high_freq_decay = 0.3  # Slower decay for higher frequencies
    power_spectrum *= np.exp(-high_freq_decay * (freqs / 100))

    # Randomize phase
    random_phases = np.exp(2j * np.pi * np.random.rand(len(freqs)))

    # Combine magnitude and phase into the frequency domain signal
    signal_freq_domain = np.sqrt(power_spectrum) * random_phases

    # Inverse FFT to generate the time-domain signal
    signal = np.fft.irfft(signal_freq_domain)

    # Normalize signal to have zero mean and unit variance
    signal = (signal - np.mean(signal)) / np.std(signal)

    return t, signal


def plot_custom_features():
    # Parameters
    duration = 2  # seconds
    sampling_rate = 1000  # Hz
    num_signals = 7  # Number of signals to generate

    t = None
    signals = []
    for i in range(num_signals):
        t, signal = generate_lfp_like_signal(duration, sampling_rate, seed=i)
        signals.append(signal + i * 5)  # Shift each signal on the y-axis

    # Plot stacked signals
    plt.figure(figsize=(12, 4))
    cmap = plt.colormaps["rainbow"]
    colors = cmap(np.linspace(0, 1, num_signals))

    ax = plt.subplot(111)
    for i, (signal, color) in enumerate(zip(signals, colors)):
        plt.plot(t, signal, color=color)

    plt.title("Custom Features", fontsize=40)
    plt.legend(loc="upper right")
    plt.tight_layout()
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    plt.xticks([])
    plt.yticks([])
    plt.gca().set_facecolor("none")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for jaxopt.perturbations."""

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import perturbations
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt import loss
from absl.testing import parameterized


def one_hot_argmax(inputs: jnp.ndarray) -> jnp.ndarray:
  """An argmax one-hot function for arbitrary shapes."""
  inputs_flat = jnp.reshape(inputs, (-1))
  flat_one_hot = jax.nn.one_hot(jnp.argmax(inputs_flat), inputs_flat.shape[0])
  return jnp.reshape(flat_one_hot, inputs.shape)


def ranks(inputs: jnp.ndarray) -> jnp.ndarray:
  return jnp.argsort(jnp.argsort(inputs, axis=-1), axis=-1)


def top_k_hots(values, k):
  n = values.shape[0]
  return jax.nn.one_hot(jnp.argsort(values)[-k:], n)


def scalar_function(inputs):
  """Example function with discrete operation and scalar output."""
  inputs_ranks = ranks(inputs)
  n = inputs.shape[-1]
  true_ranks = ranks(jnp.arange(n))
  return jnp.mean((inputs_ranks - true_ranks) ** 2)


def quad_function(inputs_ranks):
  """Example function with scalar output."""
  n = inputs_ranks.shape[-1]
  true_ranks = ranks(jnp.arange(n))
  return jnp.mean((inputs_ranks - true_ranks) ** 2)


class PerturbationsArgmaxTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()
    self.num_samples = 1000
    self.sigma = 0.5
    self.rng = jax.random.PRNGKey(0)

  def test_argmax(self):
    """Checks that the perturbed argmax approximates well the softmax."""

    theta_matrix = jnp.array([[-0.5, 0.2, 0.1],
                              [-0.2, 0.1, 0.4]])
    single_argmax = one_hot_argmax(theta_matrix[0])
    test_single_argmax = jnp.array([0., 1., 0.])
    self.assertArraysEqual(single_argmax, test_single_argmax)
    matrix_argmax = one_hot_argmax(theta_matrix)
    test_matrix_argmax = jnp.array([[0., 0., 0.],
                                    [0., 0., 1.]])
    self.assertArraysEqual(matrix_argmax, test_matrix_argmax)
    batch_argmax = jax.vmap(one_hot_argmax)(theta_matrix)
    test_batch_argmax = jnp.array([[0., 1., 0.],
                                   [0., 0., 1.]])
    self.assertArraysEqual(batch_argmax, test_batch_argmax)

    square_norm_fun = lambda x: jnp.sum(x * x, axis=-1)

    theta_batch = jnp.array([[-0.6, 1.9, -0.2, 1.1, -1.0],
                             [-0.6, 1.0, -0.2, 1.8, -1.0]], dtype=jnp.float32)
    one_hot_argmax_batch = jax.jit(jax.vmap(one_hot_argmax))(theta_batch)
    one_hot_test = jnp.array([[0.0, 1.0, 0.0, 0.0, 0.0],
                              [0.0, 0.0, 0.0, 1.0, 0.0]], dtype=jnp.float32)
    self.assertArraysEqual(one_hot_argmax_batch, one_hot_test)

    pert_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Gumbel()))

    # same as pert_argmax_fun with control variate 
    pert_argmax_fun_rv = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Gumbel(),
        control_variate=True))


    rngs_batch = jax.random.split(self.rng, 2)
    pert_argmax = jax.vmap(pert_argmax_fun)(theta_batch, rngs_batch)
    pert_argmax_rv = jax.vmap(pert_argmax_fun_rv)(theta_batch, rngs_batch)
    soft_argmax = jax.nn.softmax(theta_batch/self.sigma)
    self.assertArraysAllClose(pert_argmax, soft_argmax, atol=3e-2)
    self.assertArraysAllClose(pert_argmax_rv, soft_argmax, atol=3e-2)

    theta_batch_repeat = jnp.array([[-0.6, 1.9, -0.2, 1.1, -1.0],
                                    [-0.6, 1.9, -0.2, 1.1, -1.0]],
                                   dtype=jnp.float32)
    pert_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Gumbel()))

    # same as pert_argmax_fun but with control variate 
    pert_argmax_fun_rv = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Gumbel(),
        control_variate=True))

    rngs_batch = jax.random.split(self.rng, 2)
    pert_argmax_repeat = jax.vmap(pert_argmax_fun)(theta_batch_repeat,
                                                   rngs_batch)
    pert_argmax_repeat_rv = jax.vmap(pert_argmax_fun_rv)(theta_batch_repeat,
                                                         rngs_batch)

    self.assertArraysAllClose(pert_argmax_repeat[0], pert_argmax_repeat[1],
                              atol=2e-2)
    self.assertArraysAllClose(pert_argmax_repeat_rv[0], pert_argmax_repeat_rv[1],
                              atol=2e-2)

    delta_noise = pert_argmax_repeat[0] - pert_argmax_repeat[1]
    delta_noise_rv = pert_argmax_repeat_rv[0] - pert_argmax_repeat_rv[1]
    self.assertNotAlmostEqual(jnp.linalg.norm(delta_noise), 0)
    self.assertNotAlmostEqual(jnp.linalg.norm(delta_noise_rv), 0)

    def square_loss(theta_batch, rng):
      batch_size = theta_batch.shape[0]
      rngs_batch = jax.random.split(rng, batch_size)
      pert_argmax_batch = jax.vmap(pert_argmax_fun)(theta_batch, rngs_batch)
      return jnp.mean(square_norm_fun(pert_argmax_batch))

    # with control variate 
    def square_loss_rv(theta_batch, rng):
      batch_size = theta_batch.shape[0]
      rngs_batch = jax.random.split(rng, batch_size)
      pert_argmax_batch_rv = jax.vmap(pert_argmax_fun_rv)(theta_batch, rngs_batch)
      return jnp.mean(square_norm_fun(pert_argmax_batch_rv))


    grad_pert = jax.grad(square_loss)(theta_batch, self.rng)
    grad_pert_rv = jax.grad(square_loss_rv)(theta_batch, self.rng)

    def square_loss_soft(theta):
      soft_max = jax.nn.softmax(theta/self.sigma)
      return jnp.mean(square_norm_fun(soft_max))

    grad_soft = jax.grad(square_loss_soft)(theta_batch)
    self.assertArraysAllClose(grad_pert, grad_soft, atol=1e-1)
    self.assertArraysAllClose(grad_pert_rv, grad_soft, atol=1e-1)

    """
    Test ensuring that for small sigma, control variate indeed leads
    to a Jacobian that is closer to that of the softmax.
    """
    sigma = 0.01

    pert_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=0.01,
        noise=perturbations.Gumbel()))

    # same as pert_argmax_fun but with control variate 
    pert_argmax_fun_rv = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=0.01,
        noise=perturbations.Gumbel(),
        control_variate=True))

    rng, _  = jax.random.split(self.rng, 2)
    jac_softmax = jax.jacfwd(jax.nn.softmax)
    jac_pert_argmax_fun = jax.jacfwd(pert_argmax_fun)
    jac_pert_argmax_fun_rv = jax.jacfwd(pert_argmax_fun_rv)

    # calculate the jacobians
    js = jac_softmax(theta_matrix[0])
    jp = jac_pert_argmax_fun(theta_matrix[0], rng)
    jp_rv = jac_pert_argmax_fun_rv(theta_matrix[0], rng)

    # distance of pert argmax jacobians from softmax jacobian
    jac_dist = jnp.linalg.norm((js - jp) ** 2)
    jac_dist_rv = jnp.linalg.norm((js - jp_rv) ** 2)
    self.assertLessEqual(jac_dist_rv, jac_dist)


  def test_distrax(self):
    """Checks that the function is compatible with distrax distributions."""
    try:
      import distrax
    except ImportError:
      return

    theta_batch = jnp.array([[-0.5, 0.2, 0.1],
                             [-0.2, 0.1, 0.4]])
    pert_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    rngs_batch = jax.random.split(self.rng, 2)
    pert_argmax = jax.vmap(pert_argmax_fun)(theta_batch, rngs_batch)

    dist_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=distrax.Normal(loc=0., scale=1.)))
    dist_argmax = jax.vmap(dist_argmax_fun)(theta_batch, rngs_batch)

    self.assertArraysAllClose(pert_argmax, dist_argmax, atol=1e-6)

  def test_or(self):
    """Checks that the perturbed or returns the correct value."""
    # High value of num_samples for this specific test. Not required in normal
    # usecases, as in learning tasks.
    pert_sign = jax.jit(jax.vmap(perturbations.make_perturbed_argmax(
        argmax_fun=jnp.sign,
        num_samples=10000,
        sigma=self.sigma,
        noise=perturbations.Normal())))

    pert_sign_rv = jax.jit(jax.vmap(perturbations.make_perturbed_argmax(
        argmax_fun=jnp.sign,
        num_samples=10000,
        sigma=self.sigma,
        noise=perturbations.Normal(),
        control_variate=True)))



    theta_batch = jnp.array([[-0.5, 1.2],
                             [-0.4, -1.2],
                             [-0.2, 0.2],
                             [0.1, 0.05]])
    rngs_batch = jax.random.split(self.rng, 4)
    soft_sign = pert_sign(theta_batch, rngs_batch)
    soft_sign_rv = pert_sign_rv(theta_batch, rngs_batch)
    test_sign = 2 * jax.scipy.stats.norm.cdf(theta_batch/self.sigma) - 1
    self.assertArraysAllClose(soft_sign, test_sign, atol=1e-1)
    self.assertArraysAllClose(soft_sign_rv, test_sign, atol=1e-1)

    def pert_sum_or(x_batch, rng):
      batch_size = x_batch.shape[0]
      rngs_batch = jax.random.split(rng, batch_size)
      pert_sign_batch = pert_sign(x_batch, rngs_batch)
      return jnp.sum(jnp.max(pert_sign_batch, axis=-1))

    def pert_sum_or_rv(x_batch, rng):
      batch_size = x_batch.shape[0]
      rngs_batch = jax.random.split(rng, batch_size)
      pert_sign_batch_rv = pert_sign_rv(x_batch, rngs_batch)
      return jnp.sum(jnp.max(pert_sign_batch_rv, axis=-1))


    def explicit_sum_or(x):
      cdf_value = jax.scipy.stats.norm.cdf(x / self.sigma)
      return jnp.sum(jnp.max(2 * cdf_value - 1, axis=-1))

    grad_pert_sum = jax.grad(pert_sum_or)(theta_batch, self.rng)
    grad_pert_sum_rv = jax.grad(pert_sum_or_rv)(theta_batch, self.rng)

    grad_test = jax.grad(explicit_sum_or)(theta_batch)

    self.assertArraysAllClose(1 + grad_pert_sum, 1 + grad_test, atol=1e-1)
    self.assertArraysAllClose(1 + grad_pert_sum_rv, 1 + grad_test, atol=1e-1)

  def test_rank_small_sigma(self):

    theta = jnp.array([-0.8, 0.6, 1.2, -1.0, -0.7, 0.6, 1.1, -1.0, 0.4])
    pert_ranks_fn_small_sigma = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_samples=self.num_samples,
        sigma=1e-9,
        noise=perturbations.Normal()))

    pert_ranks_fn_small_sigma_rv = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_samples=self.num_samples,
        sigma=1e-9,
        noise=perturbations.Normal(),
        control_variate=True))

    pert_ranks_small_sigma = pert_ranks_fn_small_sigma(theta, self.rng)
    pert_ranks_small_sigma_rv = pert_ranks_fn_small_sigma_rv(theta, self.rng)
    test_ranks_no_sigma = jnp.array(ranks(theta), dtype='float32')
    self.assertArraysAllClose(pert_ranks_small_sigma,
                              test_ranks_no_sigma, atol=1e-3)
    self.assertArraysAllClose(pert_ranks_small_sigma_rv,
                              test_ranks_no_sigma, atol=1e-3)


  def test_rank_finite_diff(self):
    theta = jnp.array([-0.8, 0.6, 1.2, -1.0, -0.7, 0.6, 1.1, -1.0, 0.4])
    # High value of num_samples for this specific test. Not required in normal
    # usecases, as in learning tasks.
    pert_ranks_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_samples=100000,
        sigma=self.sigma,
        noise=perturbations.Normal()))

    square_loss_rank = lambda x, rng: jnp.mean(pert_ranks_fun(x, rng) ** 2)

    gradient_square_rank = jax.grad(square_loss_rank)(theta, self.rng)
    eps = 1e-2
    h = jax.random.uniform(self.rng, shape=theta.shape)
    rngs = jax.random.split(self.rng, 2)
    sq_loss_plus_h = square_loss_rank(theta + eps * h, rngs[0])
    sq_loss_minus_h = square_loss_rank(theta - eps * h, rngs[1])
    delta_num = (sq_loss_plus_h - sq_loss_minus_h) / (2 * eps)
    delta_lin = jnp.sum(gradient_square_rank * h)

    self.assertArraysAllClose(delta_num, delta_lin, atol=1e-1, rtol=.5)


class PerturbationsMaxTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()
    self.num_samples = 1000
    self.sigma = 0.5
    self.rng = jax.random.PRNGKey(0)
    self.theta_batch = jnp.array([[-0.6, 1.9, -0.2, 1.7, -1.0],
                                  [-0.6, 1.0, -0.2, 1.8, -1.0]],
                                 dtype=jnp.float32)

  def test_max_small_sigma(self):
    """Checks that the perturbed max is close to the max for small sigma."""
    pert_max_small_sigma_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=1e-7,
        noise=perturbations.Gumbel()))
    rngs_batch = jax.random.split(self.rng, 2)
    pert_max_small_sigma = jax.vmap(pert_max_small_sigma_fun)(self.theta_batch,
                                                              rngs_batch)

    self.assertArraysAllClose(pert_max_small_sigma,
                              jnp.array([1.9, 1.8]), atol=1e-6)

  def test_max_jensen(self):
    """Checks that the noise increases the value of the expected max."""
    rngs_batch = jax.random.split(self.rng, 2)
    pert_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=10000,
        sigma=self.sigma,
        noise=perturbations.Gumbel()))
    pert_max = jax.vmap(pert_max_fun)(self.theta_batch, rngs_batch)
    self.assertGreater(pert_max[0], 1.9)
    self.assertGreater(pert_max[1], 1.8)

  def test_grads(self):
    """Tests that the perturbed and regularized maxes match."""
    pert_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=10000,
        sigma=self.sigma,
        noise=perturbations.Gumbel()))

    def sum_max(theta_batch, rng):
      batch_size = theta_batch.shape[0]
      rngs_batch = jax.random.split(rng, batch_size)
      pert_max_batch = jax.vmap(pert_max_fun)(theta_batch, rngs_batch)
      return jnp.sum(pert_max_batch)

    soft_argmax = jax.nn.softmax(self.theta_batch/self.sigma)

    grad_pert = jax.grad(sum_max)(self.theta_batch, self.rng)

    self.assertArraysAllClose(soft_argmax, grad_pert, atol=1e-2)

  def test_noise_iid(self):
    """Checks that different elements of the batch have different noises."""
    pert_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=10000,
        sigma=self.sigma,
        noise=perturbations.Gumbel()))
    theta_batch_repeat = jnp.array([[-0.6, 1.9, -0.2, 1.1, -1.0],
                                    [-0.6, 1.9, -0.2, 1.1, -1.0]],
                                   dtype=jnp.float32)
    rngs_batch = jax.random.split(self.rng, 2)
    pert_max_repeat = jax.vmap(pert_max_fun)(theta_batch_repeat, rngs_batch)
    self.assertArraysAllClose(pert_max_repeat[0], pert_max_repeat[1],
                              atol=1e-2)
    delta_noise = pert_max_repeat[0] - pert_max_repeat[1]
    self.assertNotAlmostEqual(jnp.linalg.norm(delta_noise), 0)

  def test_distrax(self):
    """Checks that the function is compatible with distrax distributions."""
    try:
      import distrax
    except ImportError:
      return

    theta_batch = jnp.array([[-0.5, 0.2, 0.1],
                             [-0.2, 0.1, 0.4]])
    pert_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    rngs_batch = jax.random.split(self.rng, 2)
    pert_max = jax.vmap(pert_max_fun)(theta_batch, rngs_batch)

    dist_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=one_hot_argmax,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=distrax.Normal(loc=0., scale=1.)))
    dist_max = jax.vmap(dist_max_fun)(theta_batch, rngs_batch)

    self.assertArraysAllClose(pert_max, dist_max, atol=1e-6)

  def test_rank_finite_diff(self):
    theta = jnp.array([-0.8, 0.6, 1.2, -1.0, -0.7, 0.6, 1.1, -1.0, 0.4])
    # High value of num_samples for this specific test. Not required in normal
    # usecases, as in learning tasks.
    pert_ranks_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=ranks,
        num_samples=100000,
        sigma=self.sigma,
        noise=perturbations.Normal()))

    gradient_ranks_value = jax.grad(pert_ranks_fun)(theta, self.rng)
    eps = 1e-2
    h = jax.random.uniform(self.rng, shape=theta.shape)
    rngs = jax.random.split(self.rng, 2)
    delta_num = (pert_ranks_fun(theta + eps * h, rngs[0]) -
                 pert_ranks_fun(theta - eps * h, rngs[0])) / (2 * eps)
    delta_lin = jnp.sum(gradient_ranks_value * h)

    self.assertArraysAllClose(delta_num, delta_lin, rtol=1e-2)

  def test_rank_autodiff(self):
    rnd_b_s = 10
    theta_random_batch = jax.random.uniform(self.rng, (rnd_b_s, 5))
    pert_ranks_max_fun = jax.jit(perturbations.make_perturbed_max(
        argmax_fun=ranks,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    pert_ranks_argmax_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    rngs = jax.random.split(self.rng, rnd_b_s)
    autodiff_grad = jax.vmap(jax.grad(pert_ranks_max_fun))(theta_random_batch,
                                                           rngs)
    expected_grad = jax.vmap(pert_ranks_argmax_fun)(theta_random_batch, rngs)
    self.assertArraysAllClose(autodiff_grad, expected_grad, rtol=1e-6)

  def test_shape(self):
    rngs = jax.random.split(self.rng)
    values = jax.random.normal(rngs[0], (6,))

    def top_2_hots(values_k):
      values = values_k[0]
      return top_k_hots(values, 2)

    diff_top_2_hots = perturbations.make_perturbed_argmax(top_2_hots,
                                                          num_samples=100,
                                                          sigma=0.25)

    def loss_example(values, rng):
      y_true = top_k_hots(jnp.arange(6), 2)
      values_k = jnp.tile(values[jnp.newaxis, :], (2, 1))
      y_pred = diff_top_2_hots(values_k, rng)
      return jnp.sum((y_true - y_pred) ** 2)

    gradient = jax.grad(loss_example)(values, rngs[1])
    self.assertEqual(gradient.shape, values.shape)

class PerturbationsScalarTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()
    self.num_samples = 1000
    self.sigma = 0.5
    self.rng = jax.random.PRNGKey(0)
    self.theta_batch = jnp.array(
        [[-0.6, 1.9, -0.2, 1.7, -1.0], [-0.6, 1.0, -0.2, 1.8, -1.0]],
        dtype=jnp.float32,
    )
  @parameterized.product(
      control_variate=[True, False],
  )
  def test_scalar_small_sigma(self, control_variate):
    """Checks that the perturbed scalar is close to the max for small sigma."""
    pert_scalar_small_sigma_fun = jax.jit(
        perturbations.make_perturbed_fun(
            fun=scalar_function,
            num_samples=self.num_samples,
            sigma=1e-7,
            noise=perturbations.Gumbel(),
            control_variate=control_variate
        )
    )
    rngs_batch = jax.random.split(self.rng, 2)
    pert_scalar_small_sigma = jax.vmap(pert_scalar_small_sigma_fun)(
        self.theta_batch, rngs_batch
    )

    self.assertArraysAllClose(
        pert_scalar_small_sigma, jnp.array([5.2, 4.4]), atol=1e-6
    )

  @parameterized.product(
      control_variate=[True, False],
  )
  def test_grads(self, control_variate):
    """Tests that the gradients have the correct shape."""
    pert_scalar_fun = jax.jit(
        perturbations.make_perturbed_fun(
            fun=scalar_function,
            num_samples=self.num_samples,
            sigma=self.sigma,
            noise=perturbations.Gumbel(),
            control_variate=control_variate
        )
    )

    grad_pert = jax.grad(pert_scalar_fun)(self.theta_batch, self.rng)

    self.assertArraysEqual(grad_pert.shape, self.theta_batch.shape)

  @parameterized.product(
      control_variate=[True, False],
  )
  def test_noise_iid(self, control_variate):
    """Checks that different elements of the batch have different noises."""
    pert_scalar_fun = jax.jit(
        perturbations.make_perturbed_fun(
            fun=scalar_function,
            num_samples=self.num_samples,
            sigma=self.sigma,
            noise=perturbations.Gumbel(),
            control_variate=control_variate
        )
    )
    theta_batch_repeat = jnp.array([[-0.6, 1.9, -0.2, 1.1, -1.0],
                                    [-0.6, 1.9, -0.2, 1.1, -1.0]],
                                   dtype=jnp.float32)
    rngs_batch = jax.random.split(self.rng, 2)
    pert_scalar_repeat = jax.vmap(pert_scalar_fun)(theta_batch_repeat,
                                                   rngs_batch)
    self.assertArraysAllClose(pert_scalar_repeat[0],
                              pert_scalar_repeat[1],
                              atol=1e-1, rtol=1e-2)
    delta_noise = pert_scalar_repeat[0] - pert_scalar_repeat[1]
    self.assertNotAlmostEqual(jnp.linalg.norm(delta_noise), 0)

  @parameterized.product(
      control_variate=[True, False],
  )
  def test_distrax(self, control_variate):
    """Checks that the function is compatible with distrax distributions."""
    try:
      import distrax
    except ImportError:
      return

    theta_batch = jnp.array([[-0.5, 0.2, 0.1],
                             [-0.2, 0.1, 0.4]])
    pert_scalar_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=scalar_function,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    rngs_batch = jax.random.split(self.rng, 2)
    pert_scalar = jax.vmap(pert_scalar_fun)(theta_batch, rngs_batch)

    dist_scalar_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=scalar_function,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=distrax.Normal(loc=0., scale=1.),
        control_variate=control_variate))

    dist_scalar = jax.vmap(dist_scalar_fun)(theta_batch, rngs_batch)

    self.assertArraysAllClose(pert_scalar, dist_scalar, atol=1e-6)

  @parameterized.product(
      control_variate=[True, False],
  )
  def test_grad_finite_diff(self, control_variate):
    theta = jnp.array([-0.8, 0.6, 1.2, -1.0, -0.7, 0.6, 1.1, -1.0, 0.4])
    # High value of num_samples for this specific test. Not required in normal
    # usecases, as in learning tasks.
    pert_scalar_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=scalar_function,
        num_samples=100_000,
        sigma=self.sigma,
        noise=perturbations.Normal(),
        control_variate=control_variate))

    gradient_pert = jax.grad(pert_scalar_fun)(theta, self.rng)
    eps = 1e-2
    h = jax.random.uniform(self.rng, shape=theta.shape)
    rngs = jax.random.split(self.rng, 2)
    delta_num = (pert_scalar_fun(theta + eps * h, rngs[0]) -
                 pert_scalar_fun(theta - eps * h, rngs[0])) / (2 * eps)
    delta_lin = jnp.sum(gradient_pert * h)

    self.assertArraysAllClose(delta_num, delta_lin, rtol=2e-1)

  def compare_grads(self):
    """Checks composition of gradients with only one sample."""
    pert_ranks_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_samples=1,
        sigma=self.sigma))
    pert_scalar_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=scalar_function,
        num_samples=1,
        sigma=self.sigma))
    rngs = jax.random.split(self.rng, 2)
    theta = jax.random(rngs[0], shape=(5,))
    pert_ranks = pert_ranks_fun(theta, rngs[1])
    grad_quad = jax.grad(quad_function)
    grad_pert_ranks = grad_quad(pert_ranks)
    jac_fun = jax.jacobian(fun=pert_ranks_fun)
    jac_pert_ranks = jac_fun(theta, rngs[1])
    grad_scalar_fun = jax.grad(pert_scalar_fun)
    grad_scalar = grad_scalar_fun(pert_ranks, rngs[1])
    grad_compose = grad_pert_ranks @ jac_pert_ranks
    self.assertArraysAllClose(grad_scalar, grad_compose)

  def test_fenchel_young_pert(self):
    # Checks the behavior of the FL loss with perturbed optimizers.
    pert_max_one_hot = perturbations.make_perturbed_max(one_hot_argmax,
                                                        num_samples=50000,
                                                        sigma=1.)
    # Artificially high values of num_samples to test asymptotic behavior.
    fy_loss_pert = loss.make_fenchel_young_loss(pert_max_one_hot)
    rngs = jax.random.split(self.rng, 2)
    rngs_batch = jax.random.split(self.rng, 8)
    theta_true = jax.random.uniform(rngs[0], (8, 5))
    pert_argmax = perturbations.make_perturbed_argmax(argmax_fun=one_hot_argmax,
                                                      num_samples=50000,
                                                      sigma=1.)
    y_true = jax.vmap(pert_argmax)(theta_true, rngs_batch)
    theta_random = jax.random.uniform(rngs[1], (8, 5))
    y_random = jax.vmap(pert_argmax)(theta_random, rngs_batch)
    fy_min = jax.vmap(fy_loss_pert)(y_true, theta_true, rngs_batch)
    fy_random = jax.vmap(fy_loss_pert)(y_true, theta_random, rngs_batch)
    # Checks that the loss is minimized for true value of the parameters.
    self.assertGreater(fy_random[0], fy_min[0])
    self.assertGreater(jnp.mean(fy_random), jnp.mean(fy_min))
    grad_random = jax.vmap(jax.grad(fy_loss_pert, argnums=1))(y_true,
                                                              theta_random,
                                                              rngs_batch)
    # Checks that the gradient of the loss takes the correct form.
    self.assertArraysAllClose(grad_random, y_random - y_true)
    y_one_hot = jax.vmap(one_hot_argmax)(theta_true)
    int_one_hot = jnp.where(y_one_hot == 1.)[1]
    loss_one_hot = jax.vmap(fy_loss_pert)(y_one_hot, theta_random, rngs_batch)
    log_loss = jax.vmap(loss.multiclass_logistic_loss)(int_one_hot,
                                                       theta_random)
    # Checks that the FY loss associated to simple perturbed argmax is correct.
    self.assertArraysAllClose(loss_one_hot, log_loss + jnp.euler_gamma,
                              rtol=1e-2)


class PerturbationsAnyDimTest(test_util.JaxoptTestCase):

  def setUp(self):
    super().setUp()
    self.num_samples = 1000
    self.sigma = 0.5
    self.rng = jax.random.PRNGKey(0)
    self.theta_batch = jnp.array(
        [[-0.6, 1.9, -0.2, 1.7, -1.0], [-0.6, 1.0, -0.2, 1.8, -1.0]],
        dtype=jnp.float32,
    )

  def test_scalar_small_sigma(self):
    """Checks that the perturbed scalar is close to the max for small sigma."""
    pert_fun_small_sigma = jax.jit(
        perturbations.make_perturbed_fun(
            fun=ranks,
            num_samples=self.num_samples,
            sigma=1e-7,
            noise=perturbations.Gumbel(),
        )
    )
    rngs_batch = jax.random.split(self.rng, 2)
    pert_out_small_sigma = jax.vmap(pert_fun_small_sigma)(
        self.theta_batch, rngs_batch
    )
    out_no_sigma = 1. * jax.vmap(ranks)(self.theta_batch)

    self.assertArraysAllClose(
        pert_out_small_sigma, out_no_sigma, atol=1e-6
    )

  def test_grads(self):
    """Tests that the gradients have the correct shape."""
    pert_fun = jax.jit(
        perturbations.make_perturbed_fun(
            fun=ranks,
            num_samples=self.num_samples,
            sigma=self.sigma,
            noise=perturbations.Gumbel(),
        )
    )

    grad_pert = jax.jacobian(pert_fun)(self.theta_batch[0], self.rng)
    dim = self.theta_batch.shape[1]
    self.assertArraysEqual(grad_pert.shape, (dim, dim))

  def test_noise_iid(self):
    """Checks that different elements of the batch have different noises."""
    pert_fun = jax.jit(
        perturbations.make_perturbed_fun(
            fun=ranks,
            num_samples=self.num_samples,
            sigma=self.sigma,
            noise=perturbations.Gumbel(),
        )
    )
    theta_batch_repeat = jnp.array([[-0.6, 1.9, -0.2, 1.1, -1.0],
                                    [-0.6, 1.9, -0.2, 1.1, -1.0]],
                                   dtype=jnp.float32)
    rngs_batch = jax.random.split(self.rng, 2)
    pert_repeat = jax.vmap(pert_fun)(theta_batch_repeat,
                                     rngs_batch)
    self.assertArraysAllClose(pert_repeat[0], pert_repeat[1],
                              atol=5e-2, rtol=5e-2)
    delta_noise = pert_repeat[0] - pert_repeat[1]
    self.assertNotAlmostEqual(jnp.linalg.norm(delta_noise), 0)

  def test_distrax(self):
    """Checks that the function is compatible with distrax distributions."""
    try:
      import distrax
    except ImportError:
      return

    theta_batch = jnp.array([[-0.5, 0.2, 0.1],
                             [-0.2, 0.1, 0.4]])
    pert_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=ranks,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=perturbations.Normal()))
    rngs_batch = jax.random.split(self.rng, 2)
    pert_scalar = jax.vmap(pert_fun)(theta_batch, rngs_batch)

    dist_scalar_fun = jax.jit(perturbations.make_perturbed_fun(
        fun=ranks,
        num_samples=self.num_samples,
        sigma=self.sigma,
        noise=distrax.Normal(loc=0., scale=1.)))
    dist_scalar = jax.vmap(dist_scalar_fun)(theta_batch, rngs_batch)

    self.assertArraysAllClose(pert_scalar, dist_scalar, atol=1e-6)

  def compose_grads(self):
    """Checks composition of gradients."""
    pert_ranks_fun = jax.jit(perturbations.make_perturbed_argmax(
        argmax_fun=ranks,
        num_sample=1,
        sigma=self.sigma))
    def scalar_pert_inside(inputs, rng):
      pert_ranks_inputs = pert_ranks_fun(inputs, rng)
      return jnp.sum(pert_ranks_inputs ** 2)
    def quad_local(inputs):
      return jnp.sum(inputs ** 2)
    rngs = jax.random.split(self.rng, 2)
    theta = jax.random(rngs[0], shape=(5,))
    pert_ranks = pert_ranks_fun(theta, rngs[1])
    grad_quad = jax.grad(quad_local)
    grad_pert_ranks = grad_quad(pert_ranks)
    jac_fun = jax.jacobian(fun=pert_ranks_fun)
    jac_pert_ranks = jac_fun(theta, rngs[1])
    grad_scalar_fun = jax.grad(scalar_pert_inside)
    grad_scalar = grad_scalar_fun(pert_ranks, rngs[1])
    grad_compose = grad_pert_ranks @ jac_pert_ranks
    self.assertArraysAllClose(grad_scalar, grad_compose)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import BoxCDQP
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


def _cd_qp(Q, c, l, u, tol, maxiter, verbose=0):
  """Pure NumPy implementation for test purposes."""
  x = onp.zeros(Q.shape[0])

  for it in range(maxiter):
    error = 0

    for i in range(len(x)):
      g_i = onp.dot(Q[i], x) + c[i]
      h_i = Q[i, i]

      if h_i == 0:
        continue

      x_i_new = onp.clip(x[i] - g_i / h_i, l[i], u[i])
      delta_i = x_i_new - x[i]
      error += onp.abs(delta_i)
      x[i] = x_i_new

    if verbose:
      print(it + 1, error)

    if error <= tol:
      break

  return x


class CD_QP_Test(test_util.JaxoptTestCase):

  def setUp(self):
    rng = onp.random.RandomState(0)
    num_dim = 5
    M = rng.randn(num_dim, num_dim)
    self.Q = onp.dot(M, M.T)
    self.c = rng.randn(num_dim)
    self.l = rng.randn(num_dim)
    self.u = self.l + 5 * rng.rand(num_dim)
    self.params_obj = (self.Q, self.c)
    self.params_ineq = (self.l, self.u)

  def test_forward(self):
    sol_numpy = _cd_qp(self.Q, self.c, self.l, self.u,
                       tol=1e-3, maxiter=100, verbose=0)

    # Manual loop
    params = jnp.zeros_like(sol_numpy)

    cdqp = BoxCDQP()
    state = cdqp.init_state(params)

    for _ in range(5):
      params, state = cdqp.update(params, state, params_obj=self.params_obj,
                                  params_ineq=self.params_ineq)

    self.assertAllClose(state.error, 0.0)
    self.assertAllClose(params, sol_numpy)

    # Run call.
    params = jnp.zeros_like(sol_numpy)
    params, state = cdqp.run(params, params_obj=self.params_obj,
                             params_ineq=self.params_ineq)
    self.assertAllClose(state.error, 0.0)
    self.assertAllClose(params, sol_numpy)

  def test_backward(self):
    cdqp = BoxCDQP(implicit_diff=True)
    cdqp2 = BoxCDQP(implicit_diff=False)
    init_params = jnp.zeros(self.Q.shape[0])

    def wrapper(c):
      params_obj = (self.Q, c)
      return cdqp.run(init_params, params_obj=params_obj,
                      params_ineq=self.params_ineq).params

    def wrapper2(c):
      params_obj = (self.Q, c)
      return cdqp2.run(init_params, params_obj=params_obj,
                       params_ineq=self.params_ineq).params

    J = jax.jacobian(wrapper)(self.c)
    J2 = jax.jacobian(wrapper2)(self.c)
    self.assertAllClose(J, J2)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  #jax.config.update("jax_enable_x64", True)
  absltest.main()
